"""
Main click group for the gpsdio CLI.  This must be isolated from the
other commands.
"""


import itertools
import logging
from pkg_resources import iter_entry_points
import sys

import click
import cligj.plugins
import str2type.ext

import gpsdio
import gpsdio.drivers


@cligj.plugins.group(plugins=(
    ep for ep in itertools.chain(*(iter_entry_points('gpsdio.gpsdio_commands'),
                                 iter_entry_points('gpsdio.gpsdio_plugins')))))
@click.version_option(gpsdio.__version__)
@click.option('--verbose', '-v', count=True, help="Increase verbosity.")
@click.option('--quiet', '-q', count=True, help="Decrease verbosity.")
@click.option(
    '--i-drv', metavar='NAME', default=None,
    help='Specify the input driver.  Normally auto-detected from file path.',
    type=click.Choice(list(gpsdio.drivers.BaseDriver.by_name.keys()))
)
@click.option(
    '--o-drv', metavar='NAME', default=None,
    help='Specify the output driver.  Normally auto-detected from file path.',
    type=click.Choice(list(gpsdio.drivers.BaseDriver.by_name.keys()))
)
@click.option(
    '--i-cmp', metavar='NAME', default=None,
    help='Input compression format.  Normally auto-detected from file path.',
    type=click.Choice(list(gpsdio.drivers.BaseCompressionDriver.by_name.keys()))
)
@click.option(
    '--o-cmp', metavar='NAME', default=None,
    help='Output compression format.  Normally auto-detected from file path.',
    type=click.Choice(list(gpsdio.drivers.BaseCompressionDriver.by_name.keys()))
)
@click.option(
    '--i-drv-opt', '--ido', 'i_drv_opts', metavar='NAME=VAL', multiple=True,
    callback=str2type.ext.click_cb_key_val,
    help='Input driver options.  JSON values are automatically decoded.',
)
@click.option(
    '--i-cmp-opt', '--ico', 'i_cmp_opts', metavar='NAME=VAL', multiple=True,
    callback=str2type.ext.click_cb_key_val,
    help='Input compression driver options.  JSON values are automatically decoded.',
)
@click.option(
    '--o-drv-opt', '--odo', 'o_drv_opts', metavar='NAME=VAL', multiple=True,
    callback=str2type.ext.click_cb_key_val,
    help='Output driver options.  JSON values are automatically decoded.',
)
@click.option(
    '--o-cmp-opt', '--oco', 'o_cmp_opts', metavar='NAME=VAL', multiple=True,
    callback=str2type.ext.click_cb_key_val,
    help='Output compression driver options.  JSON values are automatically decoded.',
)
@click.pass_context
def main_group(ctx, i_drv, o_drv, i_cmp, o_cmp, i_drv_opts, i_cmp_opts, o_drv_opts, o_cmp_opts, verbose, quiet):
    """
    A collection of tools for working with the GPSD JSON format (or the same format in a msgpack container)
    """

    verbosity = max(10, 30 - 10 * verbose) - quiet
    logging.basicConfig(stream=sys.stderr, level=verbosity)

    # A collection of objects subcommands need access to
    ctx.obj = {
        'i_drv': i_drv,
        'o_drv': o_drv,
        'i_cmp': i_cmp,
        'o_cmp': o_cmp,
        'i_drv_opts': i_drv_opts,
        'i_cmp_opts': i_cmp_opts,
        'o_drv_opts': o_drv_opts,
        'o_cmp_opts': o_cmp_opts,
        'verbosity': verbosity
    }

    # Subcommands can do `gpsdio.open(infile, **ctx.obj['r_opts'])` instead of picking
    # individual
    r_opts = {
        'driver': ctx.obj['i_drv'],
        'compression': ctx.obj['i_cmp'],
        'do': ctx.obj['i_drv_opts'],
        'co': ctx.obj['i_cmp_opts']
    }
    w_opts = {
        'driver': ctx.obj['o_drv'],
        'compression': ctx.obj['o_cmp'],
        'do': ctx.obj['o_drv_opts'],
        'co': ctx.obj['o_cmp_opts']
    }
    ctx.obj.update(
        r_opts=r_opts,
        w_opts=w_opts
    )



"""Read and write the GPSD format.

The API mimics that of csv.DictReader/DictWriter and uses
`gpsdio.schema` to get schema details.
"""


import logging
import sys

import six

from . import drivers
from . import schema


log = logging.getLogger('gpsdio-core')
# log.setLevel(logging.DEBUG)


builtin_open = open


__all__ = ['open', 'Stream']


def open(path, mode='r', dmode=None, cmode=None, compression=None, driver=None,
         do=None, co=None, **kwargs):

    """
    Return a `Stream`() instance that is set up to read or write with the
    specified driver.

    Parameters
    ----------
    path : str
        File to be opened.
    mode : str, optional
        Mode to open both the file and driver with.
    compression : str, optional
        Read or write compressed data by specifying a compression type.
    driver : str, optional
        Read or write data with this driver.
    do : dict, optional
        Additional options to pass to the driver.
    co : dict, optional
        Additional options to pass to the compression driver.
    kwargs : **kwargs, optional
        Additional options to pass to `Stream()`.

    Returns
    -------
    Stream
        A loaded instance of stream ready for I/O operations.
    """

    if path == '-' and ('r' in mode or 'a' in mode):
        path = sys.stdin
    elif path == '-' and ('w' in mode or 'a' in mode):
        path = sys.stdout

    log.debug("Opening: %s" % path)

    do = do or {}  # Driver options
    co = co or {}  # Compression options
    dmode = dmode or mode  # Driver mode
    cmode = cmode or mode  # Compression mode

    if isinstance(compression, six.string_types):
        comp_driver = drivers.get_compression(compression)
    elif compression is None:
        try:
            comp_driver = drivers.detect_compression_type(getattr(path, 'name', path))
        except Exception:
            comp_driver = None
    else:
        comp_driver = None

    log.debug("Compression driver: %s" % comp_driver)

    if isinstance(driver, six.string_types):
        io_driver = drivers.get_driver(driver)
    else:
        io_driver = drivers.detect_file_type(getattr(path, 'name', path))

    log.debug("Driver: %s" % io_driver)

    if comp_driver:
        stream = io_driver(comp_driver(path, mode=cmode, **co), mode=dmode, **do)
    else:
        stream = io_driver(path, mode=dmode, **do)

    log.debug("Built base I/O stream: %s" % stream)

    return Stream(stream, mode=mode, **kwargs)


class Stream(object):

    def __init__(self, stream, mode='r', force_msg=False, keep_fields=True, skip_failures=False, convert=True,
                 *args, **kwargs):

        """
        Read or write a stream of AIS data.

        Parameters
        ----------
        stream : file-like object or iterable
            Expects one dictionary per iteration.
        mode : str, optional
            Determines if stream is operating in read, write, or append mode.
        force_message : bool, optional

        """

        self._stream = stream
        self.force_msg = force_msg
        self.skip_failures = skip_failures
        self._mode = mode
        self.convert = convert
        self.keep_fields = keep_fields
        self.skip_failures = skip_failures

    def __iter__(self):
        return self

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self._stream.close()

    def __repr__(self):
        return "<%s Stream '%r', mode '%s' at %s>" % (
            "closed" if self.closed else "open",
            self._stream,
            self.mode,
            hex(id(self))
        )

    @property
    def closed(self):
        return self._stream.closed

    @property
    def mode(self):
        return self._mode

    @property
    def name(self):
        return getattr(self._stream, 'name', None)

    def next(self):

        if self.mode != 'r':
            raise IOError("Stream not open for reading")
        elif self.closed:
            raise IOError("Can't operate on a closed stream")

        line = None
        try:
            loaded = line = next(self._stream)
            if self.convert:
                loaded = schema.import_msg(line, skip_failures=self.skip_failures)
                if self.force_msg:
                    loaded = schema.force_msg(loaded, keep_fields=self.keep_fields)
            return loaded
        except StopIteration:
            raise
        except Exception as e:
            if not self.skip_failures:
                import traceback
                raise Exception("%s: %s: %s\n%s" % (getattr(self._stream, 'name', 'Unknown'), type(e), e, "    " + traceback.format_exc().replace("\n", "\n    ")))
            return {"__invalid__": {"__content__": line}}

    __next__ = next

    def close(self):
        return self._stream.close()

    def writeheader(self):

        """
        Write a file header if one is needed by the container format
        """

        if self.mode == 'a':
            raise IOError("Can't write header when appending")
        elif self.mode != 'w':
            raise IOError("Stream not open for writing")
        elif self.closed:
            raise IOError("Can't operate on a closed stream")

        return self._stream.writeheader()

    def write(self, msg):

        if self.mode not in ('w', 'a'):
            raise IOError("Stream not open for writing")
        elif self.closed:
            raise IOError("Can't operate on a closed stream")

        if not self.keep_fields and 'type' in msg:
            msg = {field: val for field, val in six.iteritems(msg)
                   if field in schema.get_default_msg(msg['type'])}

        if self.convert:
            if self.force_msg:
                msg = schema.force_msg(msg, keep_fields=self.keep_fields)
            msg = schema.export_msg(msg, skip_failures=self.skip_failures)

        self._stream.write(msg)

    def writelines(self, msgs):
        for msg in msgs:
            self.write(msg)

    # csv.DictWriter compatibility, don't remove :)
    writerow = write
    writerows = writelines


def filter(stream, expressions):

    """
    A generator to filter a stream of data with boolean Pythonic expressions.
    Multiple expressions can be provided but only messages that evaluate as
    `True` for all will be yielded.

    `eval()` is used for expression evaluation but it is given a modified global
    scope that doesn't include some blacklisted items like `exec()`, `eval()`, etc.

    Example:

        >>> import gpsdio
        >>> criteria = ("type in (1, 2, 3)", "lat' in msg", "mmsi == 366268061")
        >>> with gpsdio.open('sample-data/types.msg.gz') as stream:
        ...     for msg in gpsdio.filter(stream, criteria):
        ...        # Do something

    Parameter
    ---------
    stream : iter
        An iterable producing one message per iteration.
    expressions : str or tuple
        A single expression or multiple expressions to be applied to each
        message.  Only messages that pass all filters will be yielded

    Yields
    ------
    dict
        Messages that pass all expressions.
    """

    if isinstance(expressions, six.string_types):
        expressions = expressions,

    scope_blacklist = ('eval', 'compile', 'exec', 'execfile', 'builtin', 'builtins',
                       '__builtin__', '__builtins__', 'globals', 'locals')

    global_scope = {
        k: v for k, v in globals().items() if k not in ('builtins', '__builtins__')}
    global_scope['__builtins__'] = {
        k: v for k, v in globals()['__builtins__'].items() if k not in scope_blacklist}
    global_scope['builtins'] = global_scope['__builtins__']

    for msg in stream:
        local_scope = msg.copy()
        local_scope['msg'] = msg
        for expr in expressions:
            try:
                result = eval(expr, global_scope, local_scope)
            except NameError:
                # A message doesn't contain something in the expression so just force a failure
                result = False

            if not result:
                break
        else:
            yield msg


def sort(stream, field='timestamp'):

    """
    A generator to sort data by the specified field.  Requires the entire stream
    to be held in memory.  Messages lacking the specified field are dropped.

    Parameters
    ----------
    stream : iter
        Iterator producing one message per iteration.
    field : str, optional
        Field to sort by.  Defaults to sorting by `timestamp`.
    """

    queue = six.moves.queue.PriorityQueue()
    for msg in stream:
        if field in msg:
            queue.put((msg[field], msg))

    while not queue.empty():
        yield queue.get()[1]



"""Test/sample data needed by unittests."""

import os


# Test files
TYPES_JSON_FILE = os.path.join('sample-data', 'types.json')
TYPES_JSON_GZ_FILE = os.path.join('sample-data', 'types.json.gz')
TYPES_MSG_FILE = os.path.join('sample-data', 'types.msg')
TYPES_MSG_GZ_FILE = os.path.join('sample-data', 'types.msg.gz')
TYPES_NMEA_FILE = os.path.join('sample-data', 'types.nmea')
TYPES_NMEA_GZ_FILE = os.path.join('sample-data', 'types.nmea.gz')



"""Unittests for `gpsdio`."""


from . import *

# TODO: Move this to a util file.
def compare_msg(msg1, msg2, float_tolerance=0.00001):

    for k1, k2 in zip(sorted(list(msg1)), sorted(list(msg2))):
        v1 = msg1[k1]
        v2 = msg2[k2]
        if isinstance(v1, float) and isinstance(v2, float):
            if abs(v1 - v2) > float_tolerance:
                return False
        else:
            if v1 != v2:
                return False

    return True



"""
Unittests for gpsdio.cli.main
"""


from click.testing import CliRunner

import gpsdio
import gpsdio.cli


def test_version():
    result = CliRunner().invoke(gpsdio.cli.main.main_group, [
        '--version'
    ])
    print(result.output)
    assert result.exit_code is 0
    assert gpsdio.__version__ in result.output



"""
Unittests for: `gpsdio.validate`
"""


import datetime
import json
import os.path
import random
import sys
import unittest

from click.testing import CliRunner
import six

from . import cmdtest
import gpsdio
import gpsdio.cli
import gpsdio.schema
import gpsdio.validate


class TestInfoDetails(cmdtest.CmdTest):

    def test_extend(self):
        data = {"foo": 1}
        self.assertDictEqual(data, gpsdio.validate.merge_info({}, data))
        d1 = {
            'num_rows': 3,
            'num_incomplete_rows': 1,
            'num_invalid_rows': 2,
            'lat_min': 5,
            'lat_max': 7,
            'lon_min': 11,
            'lon_max': 13,
            'min_timestamp': datetime.datetime(1970, 1, 2, 0, 0, 0),
            'max_timestamp': datetime.datetime(1971, 3, 4, 5, 6, 7),
            'is_sorted': True,
            'is_sorted_files': True,
            'mmsi_declaration': True,
            'mmsi_hist': {
                '111111111': 1,
                '222222222': 2,
            },
            'msg_type_hist': {
                1: 1,
                2: 2
            }
        }

        d2 = {
            'num_rows': 5,
            'num_incomplete_rows': 2,
            'num_invalid_rows': 3,
            'lat_min': 6,
            'lat_max': 8,
            'lon_min': 10,
            'lon_max': 12,
            'min_timestamp': datetime.datetime(1970, 1, 1, 0, 0, 0),
            'max_timestamp': datetime.datetime(1971, 2, 1, 0, 0, 0),
            'is_sorted': True,
            'is_sorted_files': True,
            'mmsi_declaration': True,
            'mmsi_hist': {
                '111111111': 1,
                '333333333': 4,
            },
            'msg_type_hist': {
                1: 4,
                3: 1
            }
        }

        expected = {
            'is_sorted': True,
            'is_sorted_files': False,
            'lat_max': 8,
            'lat_min': 5,
            'lon_max': 13,
            'lon_min': 10,
            'max_timestamp': datetime.datetime(1971, 3, 4, 5, 6, 7),
            'min_timestamp': datetime.datetime(1970, 1, 1, 0, 0),
            'mmsi_declaration': True,
            'mmsi_hist': {'111111111': 2, '222222222': 2, '333333333': 4},
            'msg_type_hist': {1: 5, 2: 2, 3: 1},
            'num_incomplete_rows': 3,
            'num_invalid_rows': 5,
            'num_rows': 8
            }

        self.assertDictEqual(expected, gpsdio.validate.merge_info(d1, d2))


class TestInfo(cmdtest.CmdTest):

    keep_tree = True
    maxDiff = None
    epoch = datetime.datetime(1970, 1, 1)

    def _random_row(self):
        msg_types = list(gpsdio.schema.fields_by_msg_type.keys())
        msg_type = msg_types[random.randint(0, len(msg_types) - 1)]

        return {
            'lon': random.uniform(-180, 180),
            'lat': random.uniform(-90, 90),
            'timestamp': datetime.datetime.fromtimestamp(
                random.randint(0, int((datetime.datetime.now() - self.epoch).total_seconds()))),
            'mmsi': random.randint(100000000, 100000010),
            'type': msg_type  # TODO: FIXME: Support all types eventually: random.randint(1, 32)
        }

    num_rows = 2
    num_invalid_rows = 1

    def setUp(self):
        cmdtest.CmdTest.setUp(self)

        # The result of _random_row has a random timestamp so this value must be overwritten when the list of rows
        # to test is computed.  Every iteration picks a new time between the last time and now.
        self.rows = []
        previous_timestamp = self.epoch
        for row in [self._random_row() for r in range(self.num_rows)]:
            now = datetime.datetime.now()
            # Since it takes far less than a second to create this set of tests rows it is possible for one of the
            # new timestamps created below to be the exact same (in seconds) as the previous timestamp.  The while
            # statement prevents that from happening
            while now == previous_timestamp:
                now = datetime.datetime.now()
            new_timestamp = datetime.datetime.utcfromtimestamp(random.randint(
                int((previous_timestamp - self.epoch).total_seconds()), int((now - self.epoch).total_seconds())))
            row['timestamp'] = new_timestamp
            previous_timestamp = new_timestamp
            self.rows.append(row)

        self.expected = {
            u'num_rows': self.num_rows + self.num_invalid_rows,
            u'num_incomplete_rows': self.num_rows,
            u'num_invalid_rows': self.num_invalid_rows,  # None of the input rows are complete messages
            u'lat_min': min([r['lat'] for r in self.rows]),
            u'lat_max': max([r['lat'] for r in self.rows]),
            u'lon_min': min([r['lon'] for r in self.rows]),
            u'lon_max': max([r['lon'] for r in self.rows]),
            u'min_timestamp': str(min([r['timestamp'] for r in self.rows]).strftime("%Y-%m-%dT%H:%M:%S.%fZ")),
            u'max_timestamp': str(max([r['timestamp'] for r in self.rows]).strftime("%Y-%m-%dT%H:%M:%S.%fZ")),
            u'mmsi_hist': {},
            u'msg_type_hist': {},
            u'is_sorted': True,
            u'is_sorted_files': True,
            u'mmsi_declaration': False
        }
        for row in self.rows:
            mmsi = str(row['mmsi'])
            if mmsi in self.expected['mmsi_hist']:
                self.expected[u'mmsi_hist'][mmsi] += 1
            else:
                self.expected[u'mmsi_hist'][mmsi] = 1

            msgtype = str(row['type'])
            if msgtype in self.expected['msg_type_hist']:
                self.expected[u'msg_type_hist'][msgtype] += 1
            else:
                self.expected[u'msg_type_hist'][msgtype] = 1

    def test_sorted(self):
        infile = os.path.join(self.dir, "rows.mmsi=123.msg")
        with open(infile, "w") as f:
            w = gpsdio.open(f, 'w')
            for row in self.rows:
                w.write(row)
            for x in six.moves.range(0, self.num_invalid_rows):
                f.write("N")

        self.expected[u'file'] = str(infile)

        result = CliRunner().invoke(gpsdio.cli.main_group, [
            'validate',
            '--print-json',
            infile
        ])
        self.assertEqual(result.exit_code, 0)

        actual = json.loads(result.output)
        self.assertDictEqual(json.loads(json.dumps(self.expected)), actual)

    def test_unsorted(self):
        self.rows[0:2] = [self.rows[1], self.rows[0]]
        infile = os.path.join(self.dir, "rows.mmsi=123.msg")
        with open(infile, "w") as f:
            w = gpsdio.open(f, 'w')
            for row in self.rows:
                w.write(row)
            for x in six.moves.range(0, self.num_invalid_rows):
                f.write("N")

        self.expected[u'file'] = str(infile)
        self.expected['is_sorted'] = False

        result = CliRunner().invoke(gpsdio.cli.main_group, [
            'validate',
            '--print-json',
            infile
        ])
        self.assertEqual(result.exit_code, 0)

        actual = json.loads(result.output)
        self.assertDictEqual(self.expected, actual)

    def test_nonjson(self):
        infile = os.path.join(self.dir, "rows.mmsi=123.msg")
        with open(infile, "w") as f:
            w = gpsdio.open(f, 'w')
            for row in self.rows:
                w.write(row)
            for x in six.moves.range(0, self.num_invalid_rows):
                f.write("N\n")

        out = self.runcmd("validate", "--verbose", "--msg-hist", "--mmsi-hist", infile).output

        self.assertIn("All rows are sorted: True", out)
        self.assertIn("->", out)


class TestValidateMessage(unittest.TestCase):

    def test_all_types(self):
        for msg_type, msg_fields in gpsdio.schema.fields_by_msg_type.items():

            # Check type field individually since the other tests force it to be correct
            assert not gpsdio.schema.validate_msg({'field': 'val'})

            for value in gpsdio.schema.CURRENT['type']['bad']:
                assert not gpsdio.schema.validate_msg(
                    {'type': value})

            # Construct a good message
            num_alterantives = max(*[len(gpsdio.schema.CURRENT[f]['good'])
                                     for f in msg_fields
                                     if f != 'type' and 'good' in gpsdio.schema.CURRENT[f]])

            for alt in xrange(0, num_alterantives):
                good_message = gpsdio.schema.get_default_msg(int(msg_type))
                good_message.update({f: gpsdio.schema.CURRENT[f]['good'][alt]
                                     for f in msg_fields
                                     if f != 'type' and 'good' in gpsdio.schema.CURRENT[f] and len(gpsdio.schema.CURRENT[f]['good']) > alt})

                assert gpsdio.schema.validate_msg(good_message), \
                    "Supposed 'good' msg failed validation: %s" % good_message

            # Creating a bad message from all of the bad values is an insufficient test because the validator
            # will start checking fields and as soon as it gets to a bad one it will flag the message as invalid.
            # Every field is checked in every message and every bad field is logged but we can't validate individual
            # fields without taking a good message and then changing one field at a time to a bad field.
            for field in msg_fields:
                if field != 'type' 'bad' in gpsdio.schema.CURRENT[field]:
                    for value in gpsdio.schema.CURRENT[field]['bad']:
                        bad_message = good_message.copy()
                        bad_message[field] = value
                        assert not gpsdio.schema.validate_msg(bad_message), \
                            "Field `%s' should have caused message to fail: %s" % (field, bad_message)



"""
Unittests for: `gpsdio._schema_def`
"""


import datetime
import unittest

import six

import gpsdio.schema
import gpsdio._schema_def


class TestDatetime2Str(unittest.TestCase):

    def test_standard(self):

        """
        Convert a datetime object to a string
        """

        now = datetime.datetime.now()

        # Manually export datetime to the expected string format
        expected_string = now.strftime(gpsdio.schema.DATETIME_FORMAT)

        # Convert the datetime object to a string with the function
        converted = gpsdio._schema_def.datetime2str(now)
        self.assertEqual(expected_string, converted)

        # Reload the string with datetime and make sure everything matches
        reloaded = datetime.datetime.strptime(converted, gpsdio.schema.DATETIME_FORMAT)
        self.assertEqual(now.year, reloaded.year)
        self.assertEqual(now.month, reloaded.month)
        self.assertEqual(now.day, reloaded.day)
        self.assertEqual(now.hour, reloaded.hour)
        self.assertEqual(now.second, reloaded.second)
        self.assertEqual(now.microsecond, reloaded.microsecond)
        self.assertEqual(now.tzinfo, reloaded.tzinfo)


class TestStr2Datetime(unittest.TestCase):

    def test_standard(self):

        """
        Convert a standard datetime string to a datetime object
        """

        string = '2014-12-19T15:29:36.479005Z'
        expected = datetime.datetime.strptime(string, gpsdio.schema.DATETIME_FORMAT)
        converted = gpsdio._schema_def.str2datetime(string)
        self.assertEqual(string, converted.strftime(gpsdio.schema.DATETIME_FORMAT))

        self.assertEqual(expected.year, converted.year)
        self.assertEqual(expected.month, converted.month)
        self.assertEqual(expected.day, converted.day)
        self.assertEqual(expected.hour, converted.hour)
        self.assertEqual(expected.second, converted.second)
        self.assertEqual(expected.tzinfo, converted.tzinfo)


def test_fields_by_msg_type():
    for msg_type, fields in six.iteritems(gpsdio._schema_def.fields_by_msg_type):
        for version, definition in six.iteritems(gpsdio._schema_def.VERSIONS):
            for f in fields:
                assert f in definition



"""
Unittests for `gpsdio.cli`
"""


import tempfile

from click.testing import CliRunner
import newlinejson as nlj

import gpsdio
import gpsdio.cli
import gpsdio.drivers
from sample_files import TYPES_MSG_GZ_FILE
from sample_files import TYPES_JSON_FILE


# Click's CliRunner() class crashes on this test because gpsdio is closing sys.stdout
# and CliRunner() expects to do that itself.  Not sure what the work-around is.
#
# def test_cat():
#     result = CliRunner().invoke(gpsdio.cli.main_group, [
#         'cat',
#         TYPES_MSG_GZ_FILE
#     ])
#     assert result.exit_code is 0
#     with gpsdio.open(TYPES_MSG_GZ_FILE) as expected:
#         for expected, actual in zip(expected, nlj.loads(result.output)):
#             assert expected == actual


def test_env():
    result = CliRunner().invoke(gpsdio.cli.main_group, [
        'env',
        '--drivers'
    ])
    assert result.exit_code is 0
    for name in gpsdio.drivers.BaseDriver.by_name.keys():
        assert name in result.output
    result = CliRunner().invoke(gpsdio.cli.main_group, [
        'env',
        '--compression'
    ])
    assert result.exit_code is 0
    for name in gpsdio.drivers.BaseCompressionDriver.by_name.keys():
        assert name in result.output

    result = CliRunner().invoke(gpsdio.cli.main_group, [
        'env',
    ])
    assert result.exit_code is not 0


def test_load():

    with open(TYPES_JSON_FILE) as f, tempfile.NamedTemporaryFile('r+') as dst:
        stdin_input = f.read()
        print(stdin_input)

        result = CliRunner().invoke(gpsdio.cli.main_group, [
            '--o-drv', 'NewlineJSON',
            '--o-cmp', 'GZIP',
            'load',
            dst.name

        ], input=stdin_input)

        dst.seek(0)
        print(result.exception)
        assert result.exit_code is 0

        with gpsdio.open(TYPES_MSG_GZ_FILE) as expected, \
                gpsdio.open(dst.name, driver='NewlineJSON', compression='GZIP') as actual:
            for e, a in zip(expected, actual):
                assert e == a


def test_etl():

    driver = 'MsgPack'
    comp = 'BZ2'

    # Process everything and sort on timestamp
    with tempfile.NamedTemporaryFile('r+') as f:
        result = CliRunner().invoke(gpsdio.cli.main_group, [
            '--o-drv', driver,
            '--o-cmp', comp,
            'etl',
            '--sort', 'timestamp',
            TYPES_MSG_GZ_FILE,
            f.name
        ])
        f.seek(0)

        assert result.exit_code is 0

        prev = None
        with gpsdio.open(f.name, driver=driver, compression=comp) as actual:
            for msg in actual:
                if prev is None:
                    prev = msg
                else:
                    assert msg['timestamp'] >= prev['timestamp']

    # Process everything and sort on mmsi
    with tempfile.NamedTemporaryFile('r+') as f:
        result = CliRunner().invoke(gpsdio.cli.main_group, [
            '--o-drv', driver,
            '--o-cmp', comp,
            'etl',
            '--sort', 'mmsi',
            TYPES_MSG_GZ_FILE,
            f.name
        ])
        f.seek(0)

        assert result.exit_code is 0

        prev = None
        with gpsdio.open(f.name, driver=driver, compression=comp) as actual:
            for msg in actual:
                if prev is None:
                    prev = msg
                else:
                    assert msg['mmsi'] >= prev['mmsi']

    # Filter and process
    with tempfile.NamedTemporaryFile('r+') as f:
        result = CliRunner().invoke(gpsdio.cli.main_group, [
            '--o-drv', driver,
            '--o-cmp', comp,
            'etl',
            '--filter', "lat and lon",
            '--sort', 'lat',
            TYPES_MSG_GZ_FILE,
            f.name
        ])
        f.seek(0)

        assert result.exit_code is 0

        prev = None
        with gpsdio.open(f.name, driver=driver, compression=comp) as actual:
            for msg in actual:
                if prev is None:
                    prev = msg
                else:
                    assert msg['lat'] >= prev['lat']



# encoding:utf-8


"""Tools for working with the GPSD format in JSON or msgpack."""


import logging
from pkg_resources import iter_entry_points


from .core import open
from .core import filter
from .core import sort
from . import schema
from . import validate


logging.basicConfig()
logger = logging.getLogger('gpsdio')


__version__ = '0.0.4'
__author__ = 'Egil Moeller, Kevin Wurster'
__email__ = 'egil@skytruth.org, kevin@skytruth.org'
__source__ = 'https://github.com/SkyTruth/gpsdio'
__license__ = """
Copyright 2014-2015 SkyTruth
Authors:
Egil Möller <egil@skytruth-org>
Kevin Wurster <kevin@skytruth.org>

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""


# Register external drivers
for ep in iter_entry_points('gpsdio.drivers'):
    try:
        ep.load()
        logger.debug("Loaded entry-point `%s'", ep.name)
    except Exception as e:
        logger.exception("Attempted to load entry-point `%s' but failed:", ep.name, exc_info=1)



"""
The schema definitions are very long and make reading `gpsdio/schema.py` difficult
"""


import datetime


DATETIME_FORMAT = "%Y-%m-%dT%H:%M:%S.%fZ"


def datetime2str(datetime_obj):

    """
    Convert a datetime object to a normalized Benthos timestamp

    Parameters
    ----------
    datetime_obj : datetime.datetime
        A loaded datetime object
    """

    if datetime_obj is None:
        return "0000-00-00T24:60:60Z"

    return datetime_obj.strftime(DATETIME_FORMAT)


def str2datetime(string):
    """
    Convert a normalized Benthos timestamp to a datetime object

    Parameters
    ----------
    string : str
        String to convert
    """

    if string == "0000-00-00T24:60:60Z":
        return None

    ms = string[20:-1]
    ms += "000000"[:6 - len(ms)]
    return datetime.datetime(
        int(string[:4]),
        int(string[5:7]),
        int(string[8:10]),
        int(string[11:13]),
        int(string[14:16]),
        int(string[17:19]),
        int(ms))


def etastr2datetime(string):
    string = string[:-1]
    dt, t = string.split("T")
    dt = dt.split("-")
    t = t.split(":")
    if '.' in t[-1]:
        t[-1], ms = t[-1].split(".")
        ms += "000000"[:6 - len(ms)]
        t.append(ms)
    dt = [int(i) for i in dt]
    t = [int(i) for i in t]
    dt = [1900, 1, 1][:-len(dt)] + dt
    t = t + [0, 0, 0, 0][len(t):]
    return datetime.datetime(*dt + t)


# Fields required by each message type, by message type ID
fields_by_msg_type = {
    1: [
        'slot_timeout', 'sync_state', 'repeat', 'lat', 'turn', 'mmsi', 'lon', 'raim', 'heading', 'scaled', 'course',
        'second', 'status', 'type', 'device', 'spare', 'maneuver', 'class', 'speed', 'accuracy'
    ],
    2: [
        'slot_timeout', 'sync_state', 'repeat', 'lat', 'turn', 'mmsi', 'lon', 'raim', 'heading', 'scaled', 'course',
        'second', 'status', 'type', 'device', 'spare', 'maneuver', 'class', 'speed', 'accuracy'
    ],
    3: [
        'slot_timeout', 'sync_state', 'repeat', 'lat', 'turn', 'mmsi', 'lon', 'raim', 'heading', 'scaled', 'course',
        'second', 'status', 'type', 'device', 'spare', 'maneuver', 'class', 'speed', 'accuracy'],
    4: [
        'slot_timeout', 'sync_state', 'repeat', 'mmsi', 'device', 'lon', 'raim', 'scaled', 'epfd', 'transmission_ctl',
        'eta', 'spare', 'slot_number', 'lat', 'type', 'class', 'accuracy'
    ],
    5: [
        'scaled', 'to_starboard', 'repeat', 'shipname', 'ais_version', 'draught', 'mmsi', 'destination', 'to_bow',
        'dte', 'to_stern', 'to_port', 'eta', 'callsign', 'imo', 'shiptype', 'device', 'spare', 'epfd', 'type', 'class'
    ],
    6: [
        'repeat', 'retransmit', 'spare2', 'dest_mmsi', 'seqno', 'mmsi', 'dac', 'ack_required', 'scaled', 'msg_seq',
        'spare', 'fid', 'device', 'type', 'class'],
    7: [
        'repeat', 'mmsi', 'scaled', 'acks', 'device', 'type', 'class'
    ],
    8: [
        'notice_type_str', 'repeat', 'dac', 'mmsi', 'link_id', 'scaled', 'duration_minutes', 'spare', 'device',
        'sub_areas', 'class', 'notice_type', 'fid', 'type'
    ],
    9: [
        'slot_timeout', 'received_stations', 'repeat', 'spare2', 'alt_sensor', 'mmsi', 'device', 'lon', 'raim', 'dte',
        'scaled', 'course', 'second', 'spare', 'lat', 'alt', 'type', 'class', 'speed', 'accuracy'
    ],
    10: [
        'scaled', 'device', 'repeat', 'spare', 'spare2', 'dest_mmsi', 'mmsi', 'type', 'class'
    ],
    11: [
        'slot_timeout', 'sync_state', 'repeat', 'mmsi', 'device', 'lon', 'raim', 'scaled', 'epfd', 'transmission_ctl',
        'eta', 'spare', 'lat', 'slot_offset', 'type', 'class', 'accuracy'
    ],
    12: [
        'repeat', 'mmsi', 'scaled', 'spare', 'device', 'class', 'retransmit', 'seqno', 'dest_mmsi', 'type'
    ],
    13: [
        'repeat', 'mmsi', 'class', 'scaled', 'device', 'mmsi4', 'mmsi3', 'mmsi2', 'mmsi1'
    ],
    14: [
        'repeat', 'text', 'mmsi', 'scaled', 'device', 'class'
    ],
    15: [
        'spare4', 'repeat', 'slot_offset_2', 'spare3', 'spare2', 'mmsi_1', 'msg_1_1', 'mmsi', 'mmsi_2', 'class',
        'scaled', 'dest_msg_1_2', 'slot_offset_1_1', 'spare', 'device', 'type', 'msg_2', 'slot_offset_1_2'
    ],
    16: [
        'repeat', 'increment1', 'increment2', 'mmsi', 'scaled', 'spare', 'device', 'dest_mmsi_b', 'dest_mmsi_a',
        'offset1', 'offset2', 'class', 'type'
    ],
    17: [
        'scaled', 'device', 'repeat', 'spare', 'spare2', 'lat', 'type', 'mmsi', 'lon', 'class'
    ],
    18: [
        'spare2', 'scaled', 'device', 'second', 'cs', 'speed', 'unit', 'lon', 'type', 'dsc', 'msg22', 'accuracy',
        'repeat', 'mmsi', 'raim', 'band', 'spare', 'lat', 'class', 'course', 'heading', 'mode', 'display'
    ],
    19: [
        'type_and_cargo', 'spare3', 'spare2', 'to_port', 'to_bow', 'scaled', 'course', 'second', 'speed',
        'to_starboard', 'lon', 'type', 'accuracy', 'repeat', 'mmsi', 'raim', 'epfd', 'spare', 'device', 'class',
        'assigned', 'to_stern', 'lat', 'shipname', 'dte', 'heading'
    ],
    20: [
        'offset4', 'offset1', 'offset2', 'offset3', 'scaled', 'increment4', 'increment3', 'increment2', 'increment1',
        'timeout3', 'timeout2', 'timeout1', 'timeout4', 'type', 'repeat', 'mmsi', 'number4', 'number2', 'number3',
        'number1', 'device', 'class', 'spare'
    ],
    21: [
        'virtual_aid', 'to_port', 'to_bow', 'scaled', 'device', 'second', 'to_starboard', 'lon', 'type', 'accuracy',
        'repeat', 'mmsi', 'raim', 'aid_type', 'spare', 'lat', 'class', 'assigned', 'to_stern', 'shipname',
        'aton_status', 'epfd', 'off_position'
    ],
    22: [
        'repeat', 'band_b', 'zonesize', 'power', 'band_a', 'mmsi', 'spare2', 'ne_lon', 'txrx', 'scaled', 'class',
        'channel_a', 'channel_b', 'device', 'spare', 'y2', 'type', 'sw_lon', 'sw_lat', 'ne_lat'
    ],
    23: [
        'repeat', 'type_and_cargo', 'spare3', 'station_type', 'interval_raw', 'mmsi', 'ne_lon', 'quiet', 'txrx',
        'scaled', 'class', 'spare', 'device', 'spare2', 'y2', 'type', 'sw_lon', 'sw_lat', 'ne_lat'
    ],
    24: [
        'scaled', 'repeat', 'shipname', 'device', 'mmsi', 'type', 'part_num', 'class'
    ],
    25: [
        'repeat', 'dest_mmsi', 'mmsi', 'structured', 'app_id', 'scaled', 'addressed', 'device', 'data', 'class'
    ],
    27: [
        'type', 'repeat', 'mmsi', 'accuracy', 'raim', 'status', 'lon', 'lat', 'speed', 'course', 'gnss'
    ]
}



# In order to ease test maintenance as outputs and inputs change the data structure below contains a test for every
# field, a value that will pass the test, and a value that will fail the test.  All information is pulled from GPSD
# (http://catb.org/gpsd/AIVDM.html) and assumes two things:
#   1. Fieldnames are unique across all messages
#   2. Fields appearing in different message types contain the same information.  For instance, the field 'sog'
#      appears in multiple types but is always speed over ground in the same units in every message.
#
# Keys are fieldnames and values are dictionaries containing three keys:
#   1. test - a function that verifies a value is acceptable for this field
#   2. good - a value that will pass the test function
#   3. bad  - a value that will not pass the test function
#
# Some fields are flags with values 0 or 1 that may be switched to Python's bool in the future but in the meantime
# we want to be positive these values are int's.  Since bool subclasses int expressions like `0 in (1, 0)` and
# `True in (0, 1)` both evaluate as `True` which could yield unexpected results.  Any test that expects an int
# also checks to make sure that int is NOT a bool, even if the field is a range and will never be bool.  Better to be
# safe here than be forced to debug some potentially ambiguous bugs elsewhere.
VERSIONS = {
    1.0: {
        'course': {
            'default': 0.0,
            'type': float,
            'units': 'degrees',
            'import': None,
            'export': None,
            'description': 'Course over ground - degrees from north',
            # TODO: Should -90 be a valid value?  Maybe `-90 < x` instead?
            'test': lambda x: 0.0 <= x < 360.0,
            'good': [45.0],
            'bad': [360.0]
        },
        'heading': {
            'default': 0,
            'type': int,
            'units': 'degrees',
            'import': None,
            'export': None,
            'description': 'True heading - degrees from north',
            'test': lambda x: 0 <= x <= 359 or x == 511,
            'good': [511],
            'bad': [-102]
        },
        'lat': {
            'default': 0.0,
            'type': float,
            'units': 'WGS84 degrees',
            'import': None,
            'export': None,
            'description': 'North/South coordinate in WGS84 degrees',
            'test': lambda x: -90 <= x <= 90 or x == 91,
            'good': [91.0],
            'bad': [-100.0]
        },
        'lon': {
            'default': 0.0,
            'type': float,
            'units': 'WGS84 degrees',
            'import': None,
            'export': None,
            'description': 'East/West coordinate in WGS84 degrees',
            # TODO: Should -180 be a valid value?  Maybe `-180 < x` instead?
            'test': lambda x: -180 <= x <= 180 or x == 181,
            'good': [181.0],
            'bad': [-180.1]
        },
        'mmsi': {
            'default': 1234567890,
            'type': int,
            'units': 'N/A',
            'import': None,
            'export': None,
            'description': 'Mobile Marine Service Identifier',
            'test': lambda x: not isinstance(x, bool),
            'good': [123],
            'bad': [True]
        },
        'timestamp': {
            'default': str2datetime('1970-01-01T00:00:00.0Z'),
            'type': datetime.datetime,
            'import': str2datetime,
            'export': datetime2str,
            'units': 'N/A',
            'description': 'Datetime format: {}'.format(DATETIME_FORMAT)
        },
        'eta': {
            'default': str2datetime('1970-01-01T00:00:00.0Z'),
            'type': datetime.datetime,
            'import': etastr2datetime,
            'export': datetime2str,
            'units': 'N/A',
            'description': 'Datetime format: {}'.format(DATETIME_FORMAT),
            'required': False
        },
        # FIXME: Confusion over type 1 / type 18 sog/speed
        'sog': {
            'default': 0.0,
            'type': float,
            'test': lambda x: 0 <= x <= 102.2 or x == 1022,
            'good': [1022],
            'bad': [103],
            'description': '',
            'units': '',
            'required': False,
            'import': None,
            'export': None
        },
        'speed': {
            'default': 0.0,
            'type': float,
            'units': 'kn/h',
            'import': None,
            'export': None,
            'description': 'Speed over ground in nautical miles per hour',
            'test': lambda x: 0 <= x <= 102.2 or x == 1022,
            'good': [1022.0],
            'bad': [103.0]
        },
        'status': {
            'default': 'Not defined',
            'type': str,
            'units': 'N/A',
            'import': None,
            'export': None,
            'description': 'Navigation status (e.g. at anchor, moored, aground, etc.)',
            # TODO: Should test if in list of allowed values
            'test': lambda x: x,
            'good': ['Moored'],
            'bad': ['']
        },
        'turn': {
            'default': None,
            'null': True,
            'type': int,
            'units': 'degrees/min',
            'import': None,
            'export': None,
            'description': 'Rate of turn',
            'test': lambda x: x in range(-127, 129),
            'good': [125],
            'bad': [-1000]
        },
        'type': {
            'default': 0,
            'type': int,
            'units': 'N/A',
            'import': None,
            'export': None,
            'description': 'NMEA message code',
            'test': lambda x: x in range(1, 28),
            'good': [5],
            'bad': [-1]
        },
        'shipname': {
            'default': '',
            'type': str,
            'units': 'N/A',
            'import': None,
            'export': None,
            'description': 'Vessel name',
            'test': lambda x: len(x) <= 2 ** 120,
            'good': ['good value'],
            'bad': [False],
            'required': False
        },
        'shiptype': {
            'default': 'Unknown',
            'type': str,
            'units': 'N/A',
            'import': None,
            'export': None,
            'description': 'Vessel type',
        },
        'callsign': {
            'default': '',
            'type': str,
            'units': 'N/A',
            'import': None,
            'export': None,
            'description': 'Vessel callsign',
            'test': lambda x: len(x) <= 2 ** 42,
            'good': ['good'],
            'bad': [123]
        },
        'destination': {
            'default': '',
            'type': str,
            'units': 'N/A',
            'import': None,
            'export': None,
            'description': 'UN/LOCODE or ERI terminal code'
        },
        'assigned': {
            'default': 0,
            'type': int,
            'units': 'N/A',
            'import': None,
            'export': None,
            'description': 'Assigned-mode flag',
            # TODO: Switch to a more Pythonic bool?
            'test': lambda x: x in (0, 1),
            'good': [1],
            'bad': [-33],
            'required': False
        },
        'to_port': {
            'default': 0,
            'type': int,
            'units': 'meters',
            'description': 'Dimension to port',
            'test': lambda x: 0 <= x <= 2 ** 6,
            'import': None,
            'export': None,
            'good': [1],
            'bad': [-34]
        },
        'accuracy': {
            'default': False,
            'description': '',
            'type': bool,
            'units': '',
            'import': None,
            'export': None,
            'good': [True],
            'bad': [2]
        },
        'ack_required': {
            'default': False,
            'description': '',
            'type': bool,
            'units': '',
            'import': None,
            'export': None
        },
        'acks': {
            'default': [],
            'description': '',
            'type': list,
            'units': '',
            'import': None,
            'export': None
        },
        'aid_type': {
            'default': 'Type of Aid to Navigation not specified',
            'description': '',
            'type': str,
            'units': '',
            'import': None,
            'export': None
        },
        'ais_version': {
            'default': 0,
            'description': '',
            'type': int,
            'units': '',
            # TODO: Should always be 0 right now.  The other vals are reserved.
            'test': lambda x: not isinstance(x, bool) and x in (0, 1, 2, 3),
            'good': [2],
            'bad': [True],
            'import': None,
            'export': None
        },
        'alt': {
            'default': 0,
            'description': '',
            'type': int,
            'units': '',
            'import': None,
            'export': None
        },
        'alt_sensor': {
            'default': 0,
            'description': '',
            'type': int,
            'units': '',
            'required': False,
            'import': None,
            'export': None
        },
        'aton_status': {
            'default': 0,
            'description': '',
            'type': int,
            'units': '',
            'required': False,
            'import': None,
            'export': None
        },
        'band': {
            'default': 1,
            'description': '',
            'type': int,
            'units': '',
            # TODO: Switch to a more Pythonic bool?
            'test': lambda x: x in (0, 1),
            'good': [0],
            'bad': [4],
            'import': None,
            'export': None
        },
        'band_a': {
            'default': 0,
            'description': '',
            'type': int,
            'units': '',
            'import': None,
            'export': None
        },
        'band_b': {
            'default': 0,
            'description': '',
            'type': int,
            'units': '',
            'import': None,
            'export': None
        },
        'channel_a': {
            'default': 0,
            'description': '',
            'type': int,
            'units': '',
            'import': None,
            'export': None
        },
        'channel_b': {
            'default': 0,
            'description': '',
            'type': int,
            'units': '',
            'import': None,
            'export': None
        },
        'class': {
            'default': 'AIS',
            'description': '',
            'type': str,
            'units': '',
            'import': None,
            'export': None
        },
        'cs': {
            'default': 1,
            'description': '',
            'type': int,
            'units': '',
            # Not bool - state
            'test': lambda x: x in (0, 1),
            'good': [0],
            'bad': [7],
            'import': None,
            'export': None
        },
        'dac': {
            'default': 1,
            'description': '',
            'type': int,
            'units': '',
            'import': None,
            'export': None
        },
        'dest_mmsi': {
            'default': 0,
            'description': '',
            'type': int,
            'units': '',
            'import': None,
            'export': None
        },
        'dest_mmsi_a': {
            'default': '',
            'description': '',
            'type': str,
            'units': '',
            'required': False,
            'import': None,
            'export': None
        },
        'dest_mmsi_b': {
            'default': '',
            'description': '',
            'type': str,
            'units': '',
            'required': False,
            'import': None,
            'export': None
        },
        'dest_msg_1_2': {
            'default': 0,
            'description': '',
            'type': int,
            'units': '',
            'required': False,
            'import': None,
            'export': None
        },
        'device': {
            'default': 'stdin',
            'description': '',
            'type': str,
            'units': '',
            'import': None,
            'export': None
        },
        'display': {
            'default': 0,
            'description': '',
            'type': int,
            'units': '',
            # Not bool - state
            'test': lambda x: x in (0, 1),
            'good': [1],
            'bad': ['j'],
            'import': None,
            'export': None
        },
        'draught': {
            'default': 0,
            'description': '',
            'type': float,
            'units': '',
            'test': lambda x: 0.0 < x <= 2 ** 8,
            'good': [1.0],
            'bad': [2 ** 8 + 1],
            'import': None,
            'export': None
        },
        'dsc': {
            'default': 1,
            'description': '',
            'type': int,
            'units': '',
            # TODO: Switch to a more Pythonic bool?
            'test': lambda x: x in (0, 1),
            'good': [1],
            'bad': [-45],
            'import': None,
            'export': None
        },
        'dte': {
            'default': 0,
            'description': '',
            'type': int,
            'units': '',
            # TODO: Switch to a more Pythonic bool if this is actually bolean and not a status
            'test': lambda x: x in (0, 1),
            'good': [0],
            'bad': [8],
            'import': None,
            'export': None
        },
        'duration_minutes': {
            'default': 0,
            'description': '',
            'type': int,
            'units': '',
            'import': None,
            'export': None
        },
        'epfd': {
            'default': 'Unknown',
            'description': '',
            'type': str,
            'units': '',
            'import': None,
            'export': None
        },
        'fid': {
            'default': 0,
            'description': '',
            'type': int,
            'units': '',
            'import': None,
            'export': None
        },
        'imo': {
            'default': 0,
            'description': '',
            'type': int,
            'units': '',
            'import': None,
            'export': None
        },
        'increment1': {
            'default': 0,
            'description': '',
            'type': int,
            'units': '',
            'import': None,
            'export': None
        },
        'increment2': {
            'default': 0,
            'description': '',
            'type': int,
            'units': '',
            'import': None,
            'export': None
        },
        'increment3': {
            'default': 0,
            'description': '',
            'type': int,
            'units': '',
            'import': None,
            'export': None
        },
        'increment4': {
            'default': 0,
            'description': '',
            'type': int,
            'units': '',
            'import': None,
            'export': None
        },
        'interval_raw': {
            'default': 11,
            'description': '',
            'type': int,
            'units': '',
            'required': False,
            'import': None,
            'export': None
        },
        'link_id': {
            'default': 0,
            'description': '',
            'type': int,
            'units': '',
            'import': None,
            'export': None
        },
        'mmsi_1': {
            'default': '',
            'description': '',
            'type': str,
            'units': '',
            'required': False,
            'import': None,
            'export': None
        },
        'mmsi_2': {
            'default': 0,
            'description': '',
            'type': int,
            'units': '',
            'required': False,
            'import': None,
            'export': None
        },
        'mmsi1': {
            'default': 0,
            'description': '',
            'type': int,
            'units': '',
            'required': False,
            'import': None,
            'export': None
        },
        'mmsi2': {
            'default': 0,
            'description': '',
            'type': int,
            'units': '',
            'required': False,
            'import': None,
            'export': None
        },
        'mmsi3': {
            'default': 0,
            'description': '',
            'type': int,
            'units': '',
            'required': False,
            'import': None,
            'export': None
        },
        'mmsi4': {
            'default': 0,
            'description': '',
            'type': int,
            'units': '',
            'required': False,
            'import': None,
            'export': None
        },
        'mode': {
            'default': False,
            'description': '',
            'type': bool,
            'units': '',
            'required': False,
            'import': None,
            'export': None
        },
        'msg22': {
            'default': 1,
            'description': '',
            'type': int,
            'units': '',
            # TODO: Switch to a more Pythonic bool?
            'test': lambda x: x in (0, 1),
            'good': [0],
            'bad': [-2],
            'required': False,
            'import': None,
            'export': None
        },
        'msg_1_1': {
            'default': 5,
            'description': '',
            'type': int,
            'units': '',
            'required': False,
            'import': None,
            'export': None
        },
        'msg_2': {
            'default': 0,
            'description': '',
            'type': int,
            'units': '',
            'required': False,
            'import': None,
            'export': None
        },
        'msg_seq': {
            'default': 0,
            'description': '',
            'type': int,
            'units': '',
            'required': False,
            'import': None,
            'export': None
        },
        'name': {
            'default': '',
            'description': 'Aid-to-Navigation name',
            'type': str,
            'units': '',
            'import': None,
            'export': None
        },
        'ne_lat': {
            'default': 0.0,
            'description': '',
            'type': float,
            # Stored as str in container for some reason
            'import': float,
            'export': str,
            'units': '',
            'import': None,
            'export': None
        },
        'ne_lon': {
            'default': 0.0,
            'description': '',
            'type': float,
            # Stored as str in container for some reason
            'import': float,
            'export': str,
            'units': '',
            'import': None,
            'export': None
        },
        'notice_type': {
            'default': 0,
            'description': '',
            'type': int,
            'units': '',
            'import': None,
            'export': None
        },
        'notice_type_str': {
            'default': 'Undefined',
            'description': '',
            'type': str,
            'units': '',
            'import': None,
            'export': None
        },
        'number1': {
            'default': 0,
            'description': '',
            'type': int,
            'units': '',
            'import': None,
            'export': None
        },
        'number2': {
            'default': 0,
            'description': '',
            'type': int,
            'units': '',
            'import': None,
            'export': None
        },
        'number3': {
            'default': 0,
            'description': '',
            'type': int,
            'units': '',
            'import': None,
            'export': None
        },
        'number4': {
            'default': 0,
            'description': '',
            'type': int,
            'units': '',
            'import': None,
            'export': None
        },
        'off_position': {
            'default': False,
            'description': '',
            'type': bool,
            'units': '',
            'import': None,
            'export': None
        },
        'offset1': {
            'default': 0,
            'description': '',
            'type': int,
            'units': '',
            'import': None,
            'export': None
        },
        'offset2': {
            'default': 0,
            'description': '',
            'type': int,
            'units': '',
            'import': None,
            'export': None
        },
        'offset3': {
            'default': 0,
            'description': '',
            'type': int,
            'units': '',
            'import': None,
            'export': None
        },
        'offset4': {
            'default': 0,
            'description': '',
            'type': int,
            'units': '',
            'import': None,
            'export': None
        },
        'part_num': {
            'default': 0,
            'description': '',
            'type': int,
            'units': '',
            'import': None,
            'export': None
        },
        'power': {
            'default': False,
            'description': '',
            'type': bool,
            'units': '',
            'import': None,
            'export': None
        },
        'quiet': {
            'default': 0,
            'description': '',
            'type': int,
            'units': '',
            'import': None,
            'export': None
        },
        'raim': {
            'default': False,
            'description': '',
            'type': bool,
            'units': '',
            'good': [False],
            'bad': [-2],
            'import': None,
            'export': None
        },
        'received_stations': {
            'default': 0,
            'description': '',
            'type': int,
            'units': '',
            'required': False,
            'import': None,
            'export': None
        },
        'repeat': {
            'default': 0,
            'description': '',
            'type': int,
            'units': '',
            'test': lambda x: 0 <= x <= 2 ** 2,
            'good': [4],
            'bad': [-1],
            'import': None,
            'export': None
        },
        'retransmit': {
            'default': True,
            'description': '',
            'type': bool,
            'units': '',
            'import': None,
            'export': None
        },
        'scaled': {
            'default': True,
            'description': '',
            'type': bool,
            'units': '',
            'import': None,
            'export': None
        },
        'second': {
            'default': 60,
            'description': '',
            'type': int,
            'units': '',
            'import': None,
            'export': None,
            'test': lambda x: x in range(0, 64),
            'good': [63],
            'bad': [64]
        },
        'seqno': {
            'default': 0,
            'description': '',
            'type': int,
            'units': '',
            'import': None,
            'export': None
        },
        'slot_number': {
            'default': 0,
            'description': '',
            'type': int,
            'units': '',
            'required': False,
            'import': None,
            'export': None
        },
        'slot_offset': {
            'default': 0,
            'description': '',
            'type': int,
            'units': '',
            'required': False,
            'import': None,
            'export': None
        },
        'slot_offset_1_1': {
            'default': 0,
            'description': '',
            'type': int,
            'units': '',
            'required': False,
            'import': None,
            'export': None
        },
        'slot_offset_1_2': {
            'default': 0,
            'description': '',
            'type': int,
            'units': '',
            'required': False,
            'import': None,
            'export': None
        },
        'slot_offset_2': {
            'default': 0,
            'description': '',
            'type': int,
            'units': '',
            'required': False,
            'import': None,
            'export': None
        },
        'slot_timeout': {
            'default': 0,
            'description': '',
            'type': int,
            'units': '',
            'required': False,
            'import': None,
            'export': None
        },
        'spare': {
            'default': 0,
            'description': '',
            'type': int,
            'units': '',
            'required': False,
            'import': None,
            'export': None
        },
        'spare2': {
            'default': 0,
            'description': '',
            'type': int,
            'units': '',
            'required': False,
            'import': None,
            'export': None

        },
        'spare3': {
            'default': 0,
            'description': '',
            'type': int,
            'units': '',
            'required': False,
            'import': None,
            'export': None

        },
        'spare4': {
            'default': 0,
            'description': '',
            'type': int,
            'units': '',
            'required': False,
            'import': None,
            'export': None

        },
        'maneuver': {
            'default': 0,
            'description': '',
            'type': int,
            'units': '',
            'test': lambda x: x in (0, 1, 2),
            'good': [2],
            'bad': [3],
            'import': None,
            'export': None
        },
        'station_type': {
            'default': 0,
            'description': '',
            'type': int,
            'units': '',
            'required': False,
            'import': None,
            'export': None
        },
        'sub_areas': {
            'default': [],
            'description': '',
            'type': list,
            'units': '',
            'import': None,
            'export': None
        },
        'sw_lon': {
            'default': 0.0,
            'description': '',
            'type': float,
            # Stored as str in container for some reason
            'import': float,
            'export': str,
            'units': '',
        },
        'sw_lat': {
            'default': 0.0,
            'description': '',
            'type': float,
            # Stored as str in container for some reason
            'import': float,
            'export': str,
            'units': ''
        },
        'sync_state': {
            'default': 0,
            'description': '',
            'type': int,
            'units': '',
            'required': False,
            'import': None,
            'export': None
        },
        'timeout1': {
            'default': 0,
            'description': '',
            'type': int,
            'units': '',
            'import': None,
            'export': None
        },
        'timeout2': {
            'default': 0,
            'description': '',
            'type': int,
            'units': '',
            'import': None,
            'export': None
        },
        'timeout3': {
            'default': 0,
            'description': '',
            'type': int,
            'units': '',
            'import': None,
            'export': None
        },
        'timeout4': {
            'default': 0,
            'description': '',
            'type': int,
            'units': '',
            'import': None,
            'export': None
        },
        'to_bow': {
            'default': 0,
            'description': '',
            'type': int,
            'units': '',
            # TODO: FIXME: This test is incorrect. value can not be > 511 according to AIS spec; check to_stern etc too
            'test': lambda x: 0 <= x <= 2 ** 9,
            'good': [1],
            'bad': [-1],
            'import': None,
            'export': None
        },
        'to_starboard': {
            'default': 0,
            'description': '',
            'type': int,
            'units': '',
            'test': lambda x: 0 <= x <= 2 ** 6,
            'good': [0],
            'bad': ['BAD'],
            'import': None,
            'export': None
        },
        'to_stern': {
            'default': 0,
            'description': '',
            'type': int,
            'units': '',
            'test': lambda x: 0 <= x <= 2 ** 9,
            'good': [0],
            'bad': [tuple],
            'import': None,
            'export': None
        },
        'transmission_ctl': {
            'default': 0,
            'description': '',
            'type': int,
            'units': '',
            'required': False,
            'import': None,
            'export': None
        },
        'txrx': {
            'default': 0,
            'description': '',
            'type': int,
            'units': '',
            'required': False,
            'import': None,
            'export': None
        },
        'type_and_cargo': {
            'default': 0,
            'description': '',
            'type': int,
            'units': '',
            'required': False,
            'import': None,
            'export': None
        },
        'unit': {
            'default': True,
            'description': '',
            'type': bool,
            'units': '',
            'required': False,
            'import': None,
            'export': None
        },
        'virtual_aid': {
            'default': False,
            'description': '',
            'type': bool,
            'units': '',
            'import': None,
            'export': None,
            'required': False
        },
        'y2': {
            'default': 0.0,
            'description': '',
            'type': float,
            'units': '',
            'required': False,
            'import': None,
            'export': None
        },
        'zonesize': {
            'default': 3,
            'description': '',
            'type': int,
            'units': '',
            'import': None,
            'export': None
        },


        # Our own columns
        'track': {
            'default': -1,
            'description': 'Track id for despoofed tracks',
            'type': int,
            'units': '',
            'import': None,
            'export': None
        },
        'measure_new_score': {
            'default': 0.0,
            'description': 'Score',
            'type': float,
            'units': '',
            'import': None,
            'export': None
        },
        'measure_coursestddev': {
            'default': 0.0,
            'description': 'Score',
            'type': float,
            'units': '',
            'import': None,
            'export': None
        },
        'measure_speedstddev': {
            'default': 0.0,
            'description': 'Score',
            'type': float,
            'units': '',
            'import': None,
            'export': None
        },
        'measure_speedavg': {
            'default': 0.0,
            'description': 'Score',
            'type': float,
            'units': '',
            'import': None,
            'export': None
        },
        # TODO: FIXME: Provide types etc for these
        'radio': {
            # TODO: What will this value be?
            'type': int,
            'good': [1],
            'bad': [False],
            'description': '',
            'units': '',
            'import': None,
            'export': None
        },
        'reserved': {
            'type': int,
            'description': '',
            'units': '',
            'import': None,
            'export': None
        },
        'regional': {
            'type': int,
            'description': '',
            'units': '',
            'import': None,
            'export': None
        },

        # Pulled from type 24 GPSD spec
        'partno': {  # TODO: FIXME: Which one of this and partnum is the correct field name?
            'type': int,
            'test': lambda x: x in (0, 1),
            'good': [0],
            'bad': [-1],
            'description': '',
            'units': '',
            'import': None,
            'export': None
        },
        'vendorid': {
            'type': str,
            'test': lambda x: len(x) <= 2 ** 18,
            'good': ['this is a gooooooooood value'],
            'bad': [int],
            'description': '',
            'units': '',
            'import': None,
            'export': None
        },
        'model': {
            'type': str,
            'test': lambda x: len(x) <= 2 ** 4,
            'good': ['something'],
            'bad': [333],
            'description': '',
            'units': '',
            'import': None,
            'export': None
        },
        'serial': {
            'type': str,
            'test': lambda x: len(x) <= 2 ** 20,
            'good': ['pawoeiras'],
            'bad': [-1],
            'description': '',
            'units': '',
            'import': None,
            'export': None
        },
        'mothership_mmsi': {
            'type': str,
            'test': lambda x: len(x) <= 2 ** 30,
            'good': ['done ... finally ...'],
            'bad': [-200],
            'description': '',
            'units': '',
            'import': None,
            'export': None
        },
        'structured': {
            'default': False,
            'type': bool,
            'description': '',
            'import': None,
            'export': None,
            'units': '',
            'required': False
        },
        'app_id': {
            'default': 0,
            'type': int,
            'description': '',
            'units': '',
            'import': None,
            'export': None
        },
        'addressed': {
            'default': False,
            'type': bool,
            'description': '',
            'units': '',
            'import': None,
            'export': None
        },
        'data': {
            'default': '',
            'type': str,
            'description': '',
            'units': '',
            'import': None,
            'export': None
        },
        'text': {
            'default': '',
            'type': str,
            'required': False,
            'description': '',
            'units': '',
            'import': None,
            'export': None
        },

        # Pulled from type 27 GPSD spec
        'gnss': {
            'default': 0,
            'type': int,
            # Not bool - state
            'test': lambda x: x in (0, 1),
            'good': [0],
            'bad': [3],
            'description': '',
            'units': '',
            'import': None,
            'export': None
        },

        # libais extensions
        'shiptype_text': {
            'default': 'Unknown',
            'type': str,
            'units': 'N/A',
            'description': 'Vessel type',
            'import': None,
            'export': None
        },
        'tagblock_timestamp': {
            'default': str2datetime('1970-01-01T00:00:00.0Z'),
            'type': datetime.datetime,
            'import': str2datetime,
            'export': datetime2str,
            'units': 'N/A',
            'description': 'Timestamp from tagblock. Datetime format: {}'.format(DATETIME_FORMAT),
        },
        'tagblock_station': {
            'default': '',
            'description': '',
            'type': str,
            'units': '',
            'import': None,
            'export': None
        }
    }
}



"""
Commandline interface for `gpsdio`.
"""


import code
import datetime
import json
import logging
import os
import sys
import warnings

import click
import six

import gpsdio
import gpsdio.drivers
import gpsdio.schema
import gpsdio.validate


def _cb_key_val(ctx, param, value):

    """
    Some options like `-ro` take `key=val` pairs that need to be transformed
    into `{'key': 'val}`.  This function can be used as a callback to handle
    all options for a specific flag, for example if the user specifies 3 reader
    options like `-ro key1=val1 -ro key2=val2 -ro key3=val3` then `click` uses
    this function to produce `{'key1': 'val1', 'key2': 'val2', 'key3': 'val3'}`.
    Parameters
    ----------
    ctx : click.Context
        Ignored
    param : click.Option
        Ignored
    value : tuple
        All collected key=val values for an option.
    Returns
    -------
    dict
    """

    output = {}
    for pair in value:
        if '=' not in pair:
            raise click.BadParameter("incorrect syntax for KEY=VAL argument: `%s'" % pair)
        else:
            key, val = pair.split('=')
            output[key] = val

    return output



@click.command()
@click.argument("infile", metavar="INPUT_FILENAME")
@click.option(
    '--print-json', is_flag=True,
    help="Print serializable JSON instead of text"
)
@click.option(
    '--msg-hist', is_flag=True, default=False,
    help="Print a type histogram"
)
@click.option(
    '--mmsi-hist', is_flag=True, default=False,
    help="Print a MMSI histogram"
)
@click.option(
    '-v', '--verbose', is_flag=True,
    help="Print details on individual messages")
@click.pass_context
def validate(ctx, infile, print_json, verbose, msg_hist, mmsi_hist):

    """
    Print info about a GPSD format AIS/GPS file
    """

    logger = logging.getLogger('gpsdio-cli-validate')
    logger.setLevel(ctx.obj['verbosity'])
    logger.debug("Starting validate")

    if os.path.isdir(infile):
        files = [os.path.join(infile, name) for name in os.listdir(infile)]
    else:
        files = [infile]

    files.sort()

    stats = {}
    for name in files:
        logger.exception("Collecting stats for {infile} ...\n".format(infile=name))

        with gpsdio.open(
                name, "r",
                driver=ctx.obj['i_drv'],
                do=ctx.obj['i_drv_opts'],
                compression=ctx.obj['i_cmp'],
                co=ctx.obj['i_cmp_opts'],
                skip_failures=True,
                force_message=False) as f:

            if verbose:
                def error_cb(type, msg, exc=None, trace=None):
                    if exc:
                        logger.exception("%s: %s: %s: %s\n" % (name, type.title(), exc, msg))
                        if trace:
                            logger.exception("%s\n" % (trace,))
                    else:
                        logger.exception("%s: %s: %s\n" % (name, type.title(), msg))
            else:
                error_cb = None
            stats = gpsdio.validate.merge_info(stats, gpsdio.validate.collect_info(f, error_cb=error_cb))

    if print_json:
        for key, value in six.iteritems(stats):
            if isinstance(value, datetime.datetime):
                stats[key] = value.strftime("%Y-%m-%dT%H:%M:%S.%fZ")
        stats['file'] = infile
        click.echo(json.dumps(stats) + "\n")
        sys.exit(0)
    else:
        click.echo("")
        click.echo("=== Report for %s ===" % infile)
        click.echo("  Number of rows: %s" % stats['num_rows'])
        click.echo("  Number of incomplete rows: %s" % stats['num_incomplete_rows'])
        click.echo("  Number of invalid rows: %s" % stats['num_invalid_rows'])
        click.echo("  All files are sorted: %s" % stats['is_sorted_files'])
        click.echo("  All rows are sorted: %s" % stats['is_sorted'])
        if stats['mmsi_declaration'] is not None:
            click.echo("  All rows match declared MMSI: %s" % stats['mmsi_declaration'])
        click.echo("  Number of unique MMSI's: %s" % len(stats['mmsi_hist']))
        click.echo("  Number of message types: %s" % len(stats['msg_type_hist']))
        click.echo("")
        click.echo("  X Min: %s" % stats['lon_min'])
        click.echo("  Y Min: %s" % stats['lat_min'])
        click.echo("  X Max: %s" % stats['lon_max'])
        click.echo("  Y Max: %s" % stats['lat_max'])
        click.echo("")
        if stats['min_timestamp'] is not None:
            _min_t = gpsdio.schema.datetime2str(stats['min_timestamp'])
        else:
            _min_t = None
        if stats['max_timestamp'] is not None:
            _max_t = gpsdio.schema.datetime2str(stats['max_timestamp'])
        else:
            _max_t = None
        click.echo("  Min timestamp: %s" % _min_t)
        click.echo("  Max timestamp: %s" % _max_t)
        if mmsi_hist:
            click.echo("")
            click.echo("  MMSI histogram:")
            for mmsi in sorted(stats['mmsi_hist'].keys()):
                click.echo("    %s -> %s" % (mmsi, stats['mmsi_hist'][mmsi]))
        if msg_hist:
            click.echo("")
            click.echo("  Message type histogram:")
            for msg_type in sorted(stats['msg_type_hist'].keys()):
                click.echo("    %s -> %s" % (msg_type, stats['msg_type_hist'][msg_type]))
        click.echo("")

    sys.exit(0)


@click.command()
@click.argument("infile", metavar="INPUT_FILENAME")
@click.argument("outfile", metavar="OUTPUT_FILENAME")
@click.pass_context
def convert(ctx, infile, outfile):

    """
    DEPRECATED - use etl instead: Converts between JSON and msgpack container formats.
    """

    deprecation_msg = "`gpsdio convert` is deprecated and will be removed before " \
                      "v1.0.  Switch to `gpsdio etl`."
    warnings.warn(deprecation_msg, DeprecationWarning, stacklevel=2)

    logger = logging.getLogger('gpsdio-cli-conver')
    logger.setLevel(ctx.obj['verbosity'])
    logging.debug('Starting convert')
    logging.warning(deprecation_msg)

    with gpsdio.open(infile) as reader:
        with gpsdio.open(outfile, 'w') as writer:
            for row in reader:
                writer.write(row)


@click.command()
@click.argument('infile', required=True)
@click.pass_context
def cat(ctx, infile):

    """
    Print messages to stdout as newline JSON.
    """

    logger = logging.getLogger('gpsdio-cli-cat')
    logger.setLevel(ctx.obj['verbosity'])
    logger.debug('Starting cat')

    with gpsdio.open(infile, **ctx.obj['r_opts']) as src:

        with gpsdio.open('-', 'w', driver='NewlineJSON', compression=False) as dst:
            for msg in src:
                dst.write(msg)


@click.command()
@click.argument('outfile', required=True)
@click.pass_context
def load(ctx, outfile):

    """
    Load newline JSON messages from stdin to a file.
    """

    logger = logging.getLogger('gpsdio-cli-load')
    logger.setLevel(ctx.obj['verbosity'])
    logger.debug('Starting load')

    with gpsdio.open('-', driver='NewlineJSON', compression=False) as src, \
            gpsdio.open(outfile, 'w', **ctx.obj['w_opts']) as dst:
        for msg in src:
            dst.write(msg)


@click.command()
@click.argument('infile', required=True)
@click.option(
    '--no-ipython', 'use_ipython', is_flag=True, default=True,
    help="Don't use IPython, even if it is available."
)
@click.pass_context
def insp(ctx, infile, use_ipython):

    # A good idea borrowed from Fiona and Rasterio

    """
    Open a dataset in an interactive inspector.

    IPython will be used if it can be imported unless otherwise specified.

    Analogous to doing:

        \b
        >>> import gpsdio
        >>> with gpsdio.open(infile) as stream:
        ...     # Operations
    """

    logger = logging.getLogger('gpsdio-cli-insp')
    logger.setLevel(ctx.obj['verbosity'])
    logger.debug('Starting insp')

    header = os.linesep.join((
        "gpsdio %s Interactive Inspector Session (Python %s)"
        % (gpsdio.__version__, '.'.join(map(str, sys.version_info[:3]))),
        "Try `help(stream)` or `next(stream)`."
    ))

    with gpsdio.open(infile, **ctx.obj['r_opts']) as src:

        scope = {
            'stream': src,
            'gpsdio': gpsdio
        }

        try:
            import IPython
        except ImportError:
            IPython = None

        if use_ipython and IPython is not None:
            IPython.embed(header=header, user_ns=scope)
        else:
            code.interact(header, local=scope)


@click.command()
@click.option(
    '--drivers', 'item', flag_value='drivers',
    help="List of registered drivers and their I/O modes."
)
@click.option(
    '--compression', 'item', flag_value='compression',
    help='List of registered compression drivers and their I/O modes.'
)
@click.pass_context
def env(ctx, item):

    """
    Information about the gpsdio environment.
    """

    logger = logging.getLogger('gpsdio-cli-env')
    logger.setLevel(ctx.obj['verbosity'])
    logger.debug('Starting env')

    if item == 'drivers':
        for name, driver in gpsdio.drivers.BaseDriver.by_name.items():
            click.echo("%s - %s" % (name, driver.io_modes))
    elif item == 'compression':
        for name, driver in gpsdio.drivers.BaseCompressionDriver.by_name.items():
            click.echo("%s - %s" % (name, driver.io_modes))
    else:
        raise click.BadParameter('A flag is required.')


@click.command()
@click.argument('infile', required=True)
@click.argument('outfile', required=True)
@click.option(
    '-f', '--filter', 'filter_expr', metavar='EXPR', multiple=True,
    help="Apply a filtering expression to the messages."
)
@click.option(
    '--sort', 'sort_field', metavar='FIELD',
    help="Sort output messages by field.  Holds the entire file in memory and drops messages "
         "lacking the specified field."
)
@click.pass_context
def etl(ctx, infile, outfile, filter_expr, sort_field):

    """
    Format conversion, filtering, and sorting.

    Data is filtered before sorting to limit the amount of data kept in memory.

    Filtering expressions take the form of Python boolean expressions and provide
    access to fields and the entire message via a custom scope.  Each field name
    can be referenced directly and the entire messages is available via a `msg`
    variable.  It is important to remember that `gpsdio` converts `timestamps` to
    `datetime.datetime()` objects internally.

    Since fields differ by message type any expression that raises a `NameError`
    when evaluated is considered a failure.

    Any Python expression that evalues as `True` or `False` can be used so so
    expressions can be combined into a single filter using `and` or split into
    multiple by using one instance of `--filter` for each side of the `and`.

    Only process messages containing a timestamp:

    \b
        $ gpsdio ${INFILE} ${OUTFILE} \\
            -f "'timestamp' in msg"

    Only process messages from May 2010 for a specific MMSI:

    \b
        $ gpsdio ${INFIE} ${OUTFILE} \\
            -f "timestamp.month == 5 and timestamp.year == 2010"" \\
            -f "mmsi == 123456789"

    Filter and sort:

    \b
        $ gpsdio ${INFILE} ${OUTFILE} \\
            -f "timestamp.year == 2010" \\
            --sort timestamp
    """

    logger = logging.getLogger('gpsdio-cli-etl')
    logger.setLevel(ctx.obj['verbosity'])
    logger.debug('Starting etl')

    with gpsdio.open(infile, **ctx.obj['r_opts']) as src:

        with gpsdio.open(outfile, 'w', **ctx.obj['w_opts']) as dst:

            iterator = gpsdio.filter(src, filter_expr) if filter_expr else src
            for msg in gpsdio.sort(iterator, sort_field) if sort_field else iterator:
                dst.write(msg)



"""
Tools for validating GPSD messages
"""


import os
import sys
import re

import six

import gpsdio.schema


def merge_info(info_a, info_b):

    """
    Joins two info dicts from info() below
    """

    def merge_hist(a, b):
        res = dict(a)
        for key, value in six.iteritems(b):
            if key in res:
                res[key] += value
            else:
                res[key] = value
        return res

    def andnone(a, b):
        if a is None:
            return b
        if b is None:
            return a
        return a and b

    def maxnone(a, b):
        if a is None:
            return b
        if b is None:
            return a
        return max(a, b)

    def minnone(a, b):
        if a is None:
            return b
        if b is None:
            return a
        return min(a, b)

    if not info_a:
        return info_b

    is_sorted = info_a['is_sorted'] and info_b['is_sorted']
    is_sorted_files = info_a['is_sorted_files'] and info_b['is_sorted_files'] and info_a['max_timestamp'] <= info_b['min_timestamp']

    return {
        'num_rows': info_a['num_rows'] + info_b['num_rows'],
        'num_incomplete_rows': info_a['num_incomplete_rows'] + info_b['num_incomplete_rows'],
        'num_invalid_rows': info_a['num_invalid_rows'] + info_b['num_invalid_rows'],
        'lat_min': minnone(info_a['lat_min'], info_b['lat_min']),
        'lat_max': maxnone(info_a['lat_max'], info_b['lat_max']),
        'lon_min': minnone(info_a['lon_min'], info_b['lon_min']),
        'lon_max': maxnone(info_a['lon_max'], info_b['lon_max']),
        'min_timestamp': minnone(info_a['min_timestamp'], info_b['min_timestamp']),
        'max_timestamp': maxnone(info_a['max_timestamp'], info_b['max_timestamp']),
        'is_sorted': is_sorted,
        'is_sorted_files': is_sorted_files,
        'mmsi_declaration': andnone(info_a['mmsi_declaration'], info_b['mmsi_declaration']),
        'mmsi_hist': merge_hist(info_a['mmsi_hist'], info_b['mmsi_hist']),
        'msg_type_hist': merge_hist(info_a['msg_type_hist'], info_b['msg_type_hist'])
        }


def collect_info(input, error_cb=None):

    """
    Get a report about a chunk of AIS data.  Report is a dictionary with the
    following keys:

        lon_min (float) -> Minimum 'lon' value
        lon_max (float) -> Maximum 'lon' value
        lat_min (float) -> Maximum 'lat' value
        lat_max (float) -> Minimum 'lat' value
        num_rows (int)   -> Total number of rows in input file
        num_invalid_rows -> Number of malformed rows
        mmsi_hist (dict) -> {mmsi: number of rows with that MMSI}
        is_sorted (bool) -> Specifies whether the entire file is sorted by timestamp
        msg_type_hist (dict) -> {message_type: number of rows with that type}
        min_timestamp (datetime) -> Earliest timestamp
        max_timestamp (datetime) -> Latest timestamp
        mmsi_declaration (bool) -> Whether an MMSI specified in the filename
            with mmsi=NUMBER is the only mmsi inside the file.

    Example output:

        {
            'num_rows': 999,
            'num_invalid_rows': 18,
            'lat_min': -52.502449999999996,
            'lat_max': 70.80154666666668,
            'lon_min': -165.8861,
            'lon_max': 153.36407833333334,
            'min_timestamp': datetime.datetime(2014, 11, 1, 7, 59, 23),
            'max_timestamp': datetime.datetime(2014, 11, 1, 7, 59, 31),
            'is_sorted': True,
            'is_sorted_files': True,
            'mmsi_declaration': None,
            'mmsi_hist': {
                '371067000': 1,
                '203999339': 1,
                '224142870': 5,
                '413505670': 2,
                '316001649': 1,
                ...
            },
            'msg_type_hist': {
                1: 416,
                2: 3,
                3: 155,
                5: 425
            }
        }

    Parameters
    ----------
    input : iterable
        An open GPSDReader or other iterable over messages
    error_cb : function(type : str, msg : dict, exc=None : exception, trace=None : str)
        Optional callback to be called for each incomplete or invalid row if given.

    Returns
    -------
    dict
    """

    stats = {
        'lon_min': None,
        'lon_max': None,
        'lat_min': None,
        'lat_max': None,
        'num_rows': 0,
        'mmsi_hist': {},
        'msg_type_hist': {},
        'num_incomplete_rows': 0,
        'num_invalid_rows': 0,
        'is_sorted': True,
        'min_timestamp': None,
        'max_timestamp': None,
        'mmsi_declaration': None
    }

    mmsi_declaration = None
    if input.name is not None and 'mmsi=' in input.name:
        mmsi_declaration = re.findall(r"mmsi=([^,.]*)[.,]", input.name)[0]
        stats['mmsi_declaration'] = True

    # Note that this is the last row that did not throw an exception on decode and is
    # not necessarily the previous row in the input
    previous_timestamp = None
    for row in input:

        # num_rows
        stats['num_rows'] += 1

        try:
            # lon_min
            if 'lon' in row and (stats['lon_min'] is None or row['lon'] < stats['lon_min']):
                stats['lon_min'] = row['lon']

            # lon_max
            if 'lon' in row and (stats['lon_max'] is None or row['lon'] > stats['lon_max']):
                stats['lon_max'] = row['lon']

            # lat_min
            if 'lat' in row and (stats['lat_min'] is None or row['lat'] < stats['lat_min']):
                stats['lat_min'] = row['lat']

            # lat_max
            if 'lat' in row and (stats['lat_max'] is None or row['lat'] > stats['lat_max']):
                stats['lat_max'] = row['lat']

            # mmsi_hist
            if 'mmsi' in row:
                if row['mmsi'] in stats['mmsi_hist']:
                    stats['mmsi_hist'][row['mmsi']] += 1
                else:
                    stats['mmsi_hist'][row['mmsi']] = 1

                if mmsi_declaration is not None:
                    if str(row['mmsi']) != mmsi_declaration:
                        stats['mmsi_declaration'] = False

            # msg_type_hist
            if 'type' in row:
                if row['type'] in stats['msg_type_hist']:
                    stats['msg_type_hist'][row['type']] += 1
                else:
                    stats['msg_type_hist'][row['type']] = 1

            # min_timestamp
            if row.get('timestamp', None) and (stats['min_timestamp'] is None or row['timestamp'] < stats['min_timestamp']):
                stats['min_timestamp'] = row['timestamp']

            # max_timestamp
            if row.get('timestamp', None) and (stats['max_timestamp'] is None or row['timestamp'] > stats['max_timestamp']):
                stats['max_timestamp'] = row['timestamp']

            # is_sorted
            # This only executes if stats['is_sorted'] = True in order to gain
            # a little optimization.  No need to test if we already know its not sorted.
            if previous_timestamp is not None and stats['is_sorted'] and row.get('timestamp', None):
                if not row['timestamp'] >= previous_timestamp:
                    stats['is_sorted'] = False

            # num_invalid_rows
            gpsdio.schema.validate_msg(row, ignore_missing=True, skip_failures=True)
            if '__invalid__' in row:
                stats['num_invalid_rows'] += 1
                if error_cb:
                    error_cb("invalid", row)
            elif not gpsdio.schema.validate_msg(row, skip_failures=True):
                stats['num_incomplete_rows'] += 1
                if error_cb:
                    error_cb("incomplete", row)

            if row.get('timestamp', None):
                previous_timestamp = row['timestamp']

        # Encountered an error - keep track of how many
        except Exception as e:
            stats['num_invalid_rows'] += 1
            if error_cb:
                import traceback                    
                error_cb("exception", row, e, traceback.format_exc(1000))

    stats['is_sorted_files'] = True

    return stats



"""Unittests for `gpsdio.core`."""


import itertools
import tempfile
import unittest

import pytest
import six

from . import compare_msg
import gpsdio
import gpsdio.schema
import gpsdio.drivers
import gpsdio.pycompat
from .sample_files import *


VALID_ROWS = [
    {'type': 5, 'shipname': 'fishy fishy', 'shiptype': 1},
    {'type': 5, 'shipname': 'fishy fishy', 'shiptype': 2},
    {'type': 5, 'shipname': 'fishy fishy', 'shiptype': 3},
    {'type': 5, 'shipname': 'fishy fishy', 'shiptype': 4}
]

INVALID_ROWS = [
    {'type': 1, 'timestamp': 'morning'},
    {'type': 1, 'timestamp': 'late'}
]


class TestOpen(unittest.TestCase):

    def setUp(self):
        self.tempfile = tempfile.NamedTemporaryFile(mode='r+')

    def tearDown(self):
        self.tempfile.close()

    def test_standard(self):

        with gpsdio.open(TYPES_MSG_FILE) as fp_msg, \
                gpsdio.open(TYPES_MSG_GZ_FILE) as fp_msg_gz, \
                gpsdio.open(TYPES_JSON_FILE) as fp_json, \
                gpsdio.open(TYPES_JSON_GZ_FILE) as fp_json_gz:

            for lines in zip(fp_msg, fp_msg_gz, fp_json, fp_json_gz):
                for pair in itertools.combinations(lines, 2):
                    self.assertTrue(compare_msg(*pair))

    def test_wrong_extension(self):

        with open(TYPES_MSG_FILE) as src:
            self.tempfile.write(src.read())
        self.tempfile.seek(0)

        with gpsdio.open(self.tempfile.name, driver='MsgPack') as actual, \
                gpsdio.open(TYPES_MSG_FILE) as expected:
            for e_line, a_line in zip(expected, actual):
                self.assertDictEqual(e_line, a_line)

        with tempfile.NamedTemporaryFile(mode='r+') as tfile:
            with open(TYPES_MSG_GZ_FILE) as src:
                tfile.write(src.read())
            tfile.seek(0)

            with gpsdio.open(tfile.name, driver='MsgPack', compression='GZIP') as actual, \
                    gpsdio.open(TYPES_MSG_GZ_FILE) as expected:
                for e_line, a_line in zip(expected, actual):
                    self.assertDictEqual(e_line, a_line)

    def test_no_detect_compression(self):

        with gpsdio.open(TYPES_MSG_FILE, compression=False) as actual, \
                gpsdio.open(TYPES_MSG_FILE) as expected:
            for e_line, a_line in zip(expected, actual):
                self.assertDictEqual(e_line, a_line)


class TestStream(unittest.TestCase):

    def setUp(self):
        self.tempfile = tempfile.NamedTemporaryFile(mode='r+')

    def tearDown(self):
        self.tempfile.close()

    def test_default_mode_is_read(self):
        with gpsdio.open(TYPES_MSG_FILE) as stream:
            self.assertEqual(stream.mode, 'r')

    def test_attrs(self):
        with gpsdio.open(TYPES_MSG_FILE) as stream:
            self.assertIsInstance(stream.__repr__(), six.string_types)
            self.assertTrue(hasattr(stream, '__next__'))

    def test_io_on_closed_stream(self):
        for mode in ('r', 'w', 'a'):
            with gpsdio.open(self.tempfile.name, mode=mode, driver='NewlineJSON') as stream:
                stream.close()
                self.assertTrue(stream.closed)
                with self.assertRaises(IOError):
                    next(stream)
                with self.assertRaises(IOError):
                    stream.write(None)
                with self.assertRaises(IOError):
                    stream.writeheader()

    def test_read_from_write_stream(self):
        with gpsdio.open(TYPES_MSG_GZ_FILE) as src, \
                gpsdio.open(self.tempfile.name, 'w', driver='NewlineJSON') as dst:
            for msg in src:
                dst.write(msg)
            with self.assertRaises(IOError):
                next(dst)

    def test_write_to_read_stream(self):
        with tempfile.NamedTemporaryFile(mode='r+') as f:
            for mode in ('r', 'a'):
                with gpsdio.open(f.name, mode=mode, driver='MsgPack') as src:
                    with self.assertRaises(IOError):
                        src.writeheader()


def test_get_driver():

    for d in [_d for _d in gpsdio.drivers.BaseDriver.by_name.values()]:
        rd = gpsdio.drivers.get_driver(d.driver_name)
        assert rd == d, "%r != %r" % (d, rd)
    try:
        gpsdio.drivers.get_driver('__---Invalid---__')
        raise TypeError("Above line should have raised a ValueError.")
    except ValueError:
        pass


def test_get_compression():

    for c in [_d for _d in gpsdio.drivers.BaseDriver.by_name if getattr(_d, 'compresion', False)]:
        cd = gpsdio.drivers.get_compression(c.name)
        assert cd == c, "%r != %r" % (c, cd)
    try:
        gpsdio.drivers.get_compression('__---Invalid---__')
        raise TypeError("Above line should have raised a ValueError.")
    except ValueError:
        pass


def test_detect_file_type():

    for d in [_d for _d in gpsdio.drivers.BaseDriver.by_name if getattr(_d, 'compresion', False)]:
        for ext in d.extensions:
            rd = gpsdio.drivers.detect_file_type('path.%s.ext' % ext)
            assert d == rd, "%r != %r" % (d, rd)
    try:
        gpsdio.drivers.detect_file_type('__---Invalid---__.ext.ext2')
        raise TypeError("Above line should have raised a ValueError.")
    except ValueError:
        pass


def test_detect_compression_type():

    for c in [_d for _d in gpsdio.drivers.BaseDriver.by_name if getattr(_d, 'compresion', False)]:
        print(c)
        print(c.extensions)
        for ext in c.extensions:
            cd = gpsdio.drivers.detect_compression_type(('path.something.%s' % ext))
            assert c == cd, "%r != %r" % (c, cd)
    try:
        gpsdio.drivers.get_compression('__---Invalid---__.ext.ext2')
        raise TypeError("Above line should have raised a ValueError.")
    except ValueError:
        pass


def test_filter():

    # Pass all through unaltered
    with gpsdio.open(TYPES_MSG_GZ_FILE) as actual, gpsdio.open(TYPES_MSG_GZ_FILE) as expected:
        for a, e in zip(gpsdio.filter(actual, "isinstance(mmsi, int)"), expected):
            assert a == e

    # Extract only types 1, 2, and 3
    passed = []
    with gpsdio.open(TYPES_MSG_GZ_FILE) as stream:
        for msg in gpsdio.filter(stream, "type in (1,2,3)"):
            passed.append(msg)
            assert msg['type'] in (1, 2, 3)
    assert len(passed) >= 3

    # Filter out everything
    with gpsdio.open(TYPES_MSG_GZ_FILE) as stream:
        for msg in gpsdio.filter(stream, ("type is 5", "mmsi is -1000")):
            assert False, "Above loop should not have executed because the filter should " \
                          "not have yielded anything."

    # Multiple expressions
    passed = []
    with gpsdio.open(TYPES_MSG_GZ_FILE) as stream:
        for msg in gpsdio.filter(
                stream, ("status == 'Under way using engine'", "mmsi is 366268061")):
            passed.append(msg)
            assert msg['mmsi'] is 366268061

    # Reference the entire message
    passed = []
    with gpsdio.open(TYPES_JSON_GZ_FILE) as stream:
        for msg in gpsdio.filter(stream, ("isinstance(msg, dict)", "'lat' in msg")):
            passed.append(msg)
            assert 'lat' in msg
    assert len(passed) >= 9

    # Multiple complex filters
    criteria = ("turn is 0 and second is 0", "mmsi == 366268061", "'lat' in msg")
    with gpsdio.open(TYPES_JSON_GZ_FILE) as stream:
        passed = [m for m in gpsdio.filter(stream, criteria)]
        assert len(passed) >= 1


# class TestBaseDriver(unittest.TestCase):
#
#     def test_attrs(self):
#         self.assertTrue(hasattr(gpsdio.drivers.BaseDriver, '__next__'))
#
#     def test_driver_context_manager(self):
#         with tempfile.NamedTemporaryFile(mode='r') as tfile:
#             with gpsdio.drivers.NewlineJSON(tfile, mode='r'):
#                 pass
#
#     def test_invalid_mode(self):
#         with self.assertRaises(ValueError):
#             gpsdio.drivers.FileDriver(None, mode='invalid')
#
#     def test_instantiated_properties(self):
#         modes = ('r', 'w', 'a')
#         name = 'drivername'
#         bd = gpsdio.drivers.BaseDriver(None, mode='r', modes=modes, name=name, reader=lambda x: x)
#         self.assertEqual(bd.modes, modes)
#         self.assertEqual(bd.name, name)



"""
Unittests for `gpsdio.drivers`.
"""


from __future__ import unicode_literals

import itertools
import tempfile

from . import compare_msg
from gpsdio import drivers
from .sample_files import *


def test_driver_read_matrix():

    open_drivers = [

        drivers.NewlineJSON(TYPES_JSON_FILE),
        drivers.NewlineJSON(open(TYPES_JSON_FILE)),

        drivers.NewlineJSON(drivers.GZIP(TYPES_JSON_GZ_FILE)),
        drivers.NewlineJSON(drivers.GZIP(open(TYPES_JSON_GZ_FILE))),

        drivers.MsgPack(TYPES_MSG_FILE),
        drivers.MsgPack(open(TYPES_MSG_FILE)),

        drivers.MsgPack(drivers.GZIP(TYPES_MSG_GZ_FILE)),
        drivers.MsgPack(drivers.GZIP(open(TYPES_MSG_GZ_FILE)))
    ]

    for lines in zip(*open_drivers):
        for pair in itertools.combinations(lines, 2):
            assert compare_msg(*pair)

    for odrv in open_drivers:
        odrv.close()

        stream = odrv.stream
        while stream:
            assert stream.closed, stream
            if hasattr(stream, '_f'):
                stream = stream._f
            else:
                stream = None


def test_msgpack():

    with tempfile.NamedTemporaryFile(mode='r+') as f:
        with drivers.MsgPack(TYPES_MSG_FILE) as src, drivers.MsgPack(f.name, 'w') as dst:
            for msg in src:
                dst.write(msg)
        f.seek(0)
        with drivers.MsgPack(TYPES_MSG_FILE) as expected, drivers.MsgPack(f.name) as actual:
            for e, a in zip(expected, actual):
                assert compare_msg(e, a)

    with tempfile.NamedTemporaryFile(mode='r+') as f:
        with drivers.MsgPack(TYPES_MSG_FILE) as src, drivers.GZIP(f.name, 'w') as gzip:
            with drivers.MsgPack(gzip, 'w') as dst:
                for msg in src:
                    dst.write(msg)
        with drivers.GZIP(f.name) as gzip:
            with drivers.MsgPack(gzip) as actual, drivers.MsgPack(TYPES_MSG_FILE) as expected:
                for e, a in zip(expected, actual):
                    assert compare_msg(e, a)



import shutil
import sys
import tempfile
import unittest

import click.testing

import gpsdio.cli


class CmdTest(unittest.TestCase):
    keep_tree = False

    def setUp(self):
        self.dir = tempfile.mkdtemp()
        self.runner = click.testing.CliRunner()
        if self.keep_tree:
            print("CmdTest is running in %s" % self.dir)

    def tearDown(self):
        if not self.keep_tree:
            shutil.rmtree(self.dir)

    def runcmd(self, *args):
        stderr = sys.stderr

        main = gpsdio.cli.main_group

        # def wrapper(*arg, **kw):
        #     clickstderr = sys.stderr
        #     sys.stderr = stderr
        #     try:
        #         return main(*arg, **kw)
        #     finally:
        #         sys.stderr = clickstderr
        # gpsdio.cli.main_group = wrapper

        try:
            return self.runner.invoke(gpsdio.cli.main_group, args, catch_exceptions=False)
        finally:
            gpsdio.cli.main_group = main



#! /usr/bin/python


"""
Setup script for gpsdio
"""


import sys

from setuptools.command.test import test as TestCommand
from setuptools import find_packages
from setuptools import setup


# and https://pytest.org/latest/goodpractises.html
class PyTest(TestCommand):
    user_options = [('pytest-args=', 'a', "Arguments to pass to py.test")]

    def initialize_options(self):
        TestCommand.initialize_options(self)
        self.pytest_args = []

    def finalize_options(self):
        TestCommand.finalize_options(self)
        self.test_args = []
        self.test_suite = True

    def run_tests(self):
        #import here, cause outside the eggs aren't loaded
        import pytest
        errno = pytest.main(self.pytest_args)
        sys.exit(errno)


version = None
author = None
email = None
source = None
with open('gpsdio/__init__.py') as f:
    for line in f:
        if line.strip().startswith('__version__'):
            version = line.split('=')[1].strip().replace('"', '').replace("'", '')
        elif line.strip().startswith('__author__'):
            author = line.split('=')[1].strip().replace('"', '').replace("'", '')
        elif line.strip().startswith('__email__'):
            email = line.split('=')[1].strip().replace('"', '').replace("'", '')
        elif line.strip().startswith('__source__'):
            source = line.split('=')[1].strip().replace('"', '').replace("'", '')
        elif None not in (version, author, email, source):
            break


setup(
    name="gpsdio",
    cmdclass={'test': PyTest},
    description="A library and command line tool to read, write and validate "
                "AIS and GPS messages in the GPSD JSON format (or the same format in a msgpack container).",
    keywords="gpsd",
    install_requires=[
        'click',
        'msgpack-python',
        'newlinejson',
        'python-dateutil',
        'six',
        'ujson',
        'cligj>=0.2',
        'str2type>=0.4'
    ],
    extras_require={
        'test': [
            'pytest',
            'pytest-cov',
        ]
    },
    version=version,
    author=author,
    author_email=email,
    license="GPL",
    url=source,
    include_package_data=True,
    packages=find_packages(),
    entry_points='''
        [console_scripts]
        gpsdio=gpsdio.cli:main_group

        [gpsdio.gpsdio_commands]
        cat=gpsdio.cli.commands:cat
        convert=gpsdio.cli.commands:convert
        env=gpsdio.cli.commands:env
        etl=gpsdio.cli.commands:etl
        insp=gpsdio.cli.commands:insp
        load=gpsdio.cli.commands:load
        validate=gpsdio.cli.commands:validate
    '''
)



import json
import os.path
import subprocess
import urllib
import pytest

from . import cmdtest

testdir = os.path.dirname(__file__)

@pytest.mark.skipif(True, reason='gpsd library is required build from source is not fully implemented')
class TestGpsdCompatibility(cmdtest.CmdTest):
    def setUp(self):
        cmdtest.CmdTest.setUp(self)

        self.utilsdir = os.path.join(testdir, "..", "utils")
        cmd = [os.path.join(self.utilsdir, "install-latest-gpsd")]
        self.gpsddir = subprocess.check_output(cmd, cwd=self.utilsdir).strip()

    def download(self, filename, url):
        with open(os.path.join(self.dir, filename), "w") as outf:
            inf = urllib.urlopen(url)
            outf.write(inf.read())
            inf.close()
        
    def validate_file(self, infile, parser):
        outfile = os.path.join(self.dir, "out.json")

        if parser == 'gpsdecode':
            cmd = "LD_LIBRARY_PATH=\"$LD_LIBRARY_PATH:%s\" %s/gpsdecode < '%s' > '%s'" % (self.gpsddir, self.gpsddir, infile, outfile)
        elif parser == 'aisdecode':
            cmd = "aisdecode --gpsd --allowUnknown --allow_missing_timestamps --encoding=json < '%s' > '%s'" % (infile, outfile)

        os.system(cmd)

        report = json.loads(self.runcmd("validate", "--print-json", "--verbose", outfile).output)

        self.assertEqual(report['num_invalid_rows'], 0)
        self.assertEqual(report['num_incomplete_rows'], 0)

    def test_gpsdecode_known_types(self):
        self.validate_file(os.path.join(os.path.dirname(__file__), "types.nmea"), 'gpsdecode')

    def test_gpsdecode_samples_from_gpsd(self):
        self.download("gpsd.nmea", "http://git.savannah.gnu.org/cgit/gpsd.git/plain/test/sample.aivdm")
        self.validate_file(os.path.join(self.dir, "gpsd.nmea"), 'gpsdecode')

    def test_gpsdecode_samples_from_libais(self):
        self.download("libais.nmea", "https://raw.githubusercontent.com/schwehr/libais/master/test/typeexamples.nmea")
        self.validate_file(os.path.join(self.dir, "libais.nmea"), 'gpsdecode')



    def test_aisdecode_known_types(self):
        self.validate_file(os.path.join(os.path.dirname(__file__), "types.nmea"), 'aisdecode')

    def test_aisdecode_samples_from_gpsd(self):
        self.download("gpsd.nmea", "http://git.savannah.gnu.org/cgit/gpsd.git/plain/test/sample.aivdm")
        self.validate_file(os.path.join(self.dir, "gpsd.nmea"), 'aisdecode')

    def test_aisdecode_samples_from_libais(self):
        self.download("libais.nmea", "https://raw.githubusercontent.com/schwehr/libais/master/test/typeexamples.nmea")
        self.validate_file(os.path.join(self.dir, "libais.nmea"), 'aisdecode')



"""
Unittests for: `gpsdio.schema`
"""


import datetime
import unittest

import gpsdio.schema


class TestCurrentSchema(unittest.TestCase):

    def test_requirements(self):
        # Make sure the current schema contains required fields

        # Helps to make sure that when a new field definition is added it actually
        # has the required elements

        required = {'default', 'type', 'units', 'description', 'import', 'export'}
        for field, definition in gpsdio.schema.CURRENT.items():
            req = required.copy()
            if definition.get('required', True):
                req.remove('default')
            self.assertEqual(req - set(definition.keys()), set(), "Field schema %s has missing fields: %s"
                             % (field, req - set(definition.keys())))
            self.assertIsInstance(definition['type'], type)
            self.assertIsInstance(definition['description'], str)
            self.assertIsInstance(definition['units'], str)
            if definition['import'] is not None:
                self.assertTrue(hasattr(definition['import'], '__call__'))
            if definition['export'] is not None:
                self.assertTrue(hasattr(definition['export'], '__call__'))


class TestValidateMessage(unittest.TestCase):

    def test_invalid(self):
        self.assertFalse(
            gpsdio.schema.validate_msg(
                {'type': 5,
                 'shipname': None,
                 }))

    def test_invalid_type(self):
        msg = gpsdio.schema.get_default_msg(1)
        self.assertTrue(gpsdio.schema.validate_msg(msg))
        msg['lat'] = 'foo'
        self.assertFalse(gpsdio.schema.validate_msg(msg))

    def test_valid(self):
        self.assertTrue(
            gpsdio.schema.validate_msg(
                gpsdio.schema.get_default_msg(1)))

    def test_error(self):
        self.assertFalse(gpsdio.schema.validate_msg(None))


class TestForce_Msg(unittest.TestCase):

    def setUp(self):
        self.row = {
            'type': 5,
            'shipname': 'SS Test Ship',
            'shiptype': 'I fish for stuff'
        }
        self.expected = self.row.copy()
        for field in gpsdio.schema.fields_by_msg_type[self.row['type']]:
            if field not in self.row:
                self.expected[field] = gpsdio.schema.CURRENT[field]['default']

    def test_standard(self):
        # Row with a valid type

        actual = gpsdio.schema.force_msg(self.row)
        self.assertDictEqual(self.expected, actual)
        for field, val in self.row.items():
            self.assertEqual(actual[field], val)

    def test_drop_unrecognized_field(self):
        # Row with a valid type but has an unrecognized field that should be dropped

        row = self.row.copy()
        row['NEW_FIELD'] = 'INVALID'
        actual = gpsdio.schema.force_msg(self.row)
        self.assertEqual(self.expected, actual)

    def test_keep_unrecognized_field(self):
        # Row with a valid type but has an unrecognized field that should be kept

        row = self.row.copy()
        row['NEW_FIELD'] = 'KEEP'

        expected = self.expected.copy()
        expected['NEW_FIELD'] = 'KEEP'

        # Explicitly define fields
        actual = gpsdio.schema.force_msg(self.row, keep_fields=['NEW_FIELD'])
        self.assertEqual(self.expected, actual)

        # Keep all fields
        actual = gpsdio.schema.force_msg(self.row, keep_fields=True)
        self.assertEqual(self.expected, actual)

    def test_invalid_type(self):
        # Row with an invalid AIS message type should raise an exception

        self.assertRaises(ValueError, gpsdio.schema.force_msg, {'type': ''})


class TestCastRow(unittest.TestCase):

    def test_standard(self):
        row = gpsdio.schema.import_msg({
                'type': 1,
                'shipname': 'SS Test Ship',
                'timestamp': '1970-01-02T03:04:05Z'
                })
        self.assertEqual(row['type'], 1)
        self.assertEqual(row['shipname'], 'SS Test Ship')
        self.assertEqual(row['timestamp'], datetime.datetime(1970, 1, 2, 3, 4, 5))

    def test_invalid_throw(self):
        self.assertRaises(
            Exception,
            gpsdio.schema.import_msg,
            {
                'type': '1',
                'shipname': 'SS Test Ship',
                'timestamp': 'late'
                })

    def test_invalid_convert(self):
        row = gpsdio.schema.import_msg(
            {
                'type': 1,
                'shipname': 'SS Test Ship',
                'timestamp': 'late'
            }, skip_failures=True
        )
        self.assertEqual(row['type'], 1)
        self.assertEqual(row['shipname'], 'SS Test Ship')
        self.assertNotIn('timestamp', row)
        self.assertIn('__invalid__', row)
        self.assertIn('timestamp', row['__invalid__'])


class TestGetMessageDefault(unittest.TestCase):

    def test_get_default_messages(self):

        for msg_type, fields in gpsdio.schema.fields_by_msg_type.items():
            actual = gpsdio.schema.get_default_msg(msg_type)
            for field in fields:
                if field == 'type':
                    continue
                if not gpsdio.schema.CURRENT[field].get('required', True):
                    continue
                self.assertIn(field, actual)
                self.assertEqual(actual[field], gpsdio.schema.CURRENT[field]['default'])

    def test_invalid_msg_type(self):
        # Try to get a row for a message type that does not exist

        msg_types = [None, -1, -1.23, 10000000, '', [], {}]
        for mt in msg_types:
            self.assertRaises(ValueError, gpsdio.schema.get_default_msg, mt)


class TestExportRow(unittest.TestCase):

    def test_standard(self):
        datetime_now = datetime.datetime.now()
        row = gpsdio.schema.export_msg(
            {
                'type': 1,
                'shipname': 'SS Test Ship',
                'speed': 3.2,
                'timestamp': datetime_now
            }
        )
        self.assertEqual(row['type'], 1)
        self.assertEqual(row['shipname'], 'SS Test Ship')
        self.assertEqual(row['speed'], 3.2)
        self.assertEqual(row['timestamp'], datetime_now.strftime(gpsdio.schema.DATETIME_FORMAT))

    def test_invalid_throw(self):
        with self.assertRaises(Exception):
            gpsdio.schema.export_msg(
                {
                    'type': '1',
                    'shipname': 'SS Test Ship',
                    'timestamp': 'late'
                }
            )

    def test_invalid_convert(self):
        row = gpsdio.schema.export_msg(
            {
                'type': 1,
                'shipname': 'SS Test Ship',
                'timestamp': 'late'
            }, skip_failures=True
        )
        self.assertEqual(row['type'], 1)
        self.assertEqual(row['shipname'], 'SS Test Ship')
        self.assertNotIn('timestamp', row)
        self.assertIn('__invalid__', row)
        self.assertIn('timestamp', row['__invalid__'])


class TestFieldDefaults(unittest.TestCase):

    def test_field_check(self):
        # Make sure all default fields are defined in the schema

        for fields in gpsdio.schema.fields_by_msg_type.values():
            for f in fields:
                self.assertTrue(f in gpsdio.schema.CURRENT, msg="Field {} not found in schema".format(f))



#!/usr/bin/env python

"""
Drivers for reading and writing a variety of formats and compression.
"""


import bz2
from codecs import open as codecs_open
import gzip

import msgpack
import newlinejson
import six
import ujson


newlinejson.JSON = ujson


def get_driver(name):

    """
    Accepts a string and returns the driver class associated with that name.
    """

    if name in BaseDriver.by_name:
        return BaseDriver.by_name[name]
    else:
        raise ValueError("Unrecognized driver name: %s" % name)


def get_compression(name):

    """
    Accepts a string and returns the driver class associated with that name.
    """

    if name in BaseCompressionDriver.by_name:
        return BaseCompressionDriver.by_name[name]
    else:
        raise ValueError("Unrecognized compression name: %s" % name)


def detect_file_type(path):

    """
    Given an file path, attempt to determine appropriate driver.
    """

    for ext in path.split('.')[-2:]:
        if ext in BaseDriver.by_extension:
            return BaseDriver.by_extension[ext]
    else:
        raise ValueError("Can't determine driver: %s" % path)


def detect_compression_type(path):

    """
    Given a file path, attempt to determine the appropriate compression driver.
    """

    ext = path.rpartition('.')[-1]
    if ext in BaseCompressionDriver.by_extension:
        return BaseCompressionDriver.by_extension[ext]
    else:
        raise ValueError("Can't determine compression: %s" % path)


class _RegisterDriver(type):

    """
    Keep track of drivers, their names, and extensions for easy retrieval later.
    """

    def __init__(driver, name, bases, members):
        # TODO: Add validation.  What methods are required?
        """
        Register drivers by name in one dictionary and by extension in another.
        """

        type.__init__(driver, name, bases, members)
        if members.get('register', True):
            driver.by_name[driver.driver_name] = driver
            for ext in driver.extensions:
                driver.by_extension[ext] = driver


class BaseDriver(six.with_metaclass(_RegisterDriver, object)):

    """
    Provides driver registration and the baseline methods required for driver
    operation.  All other non-compression drivers must subclass this class if
    they want to be registered.  Compression drivers should subclass
    `BaseCompressionDriver()`.


    Creating a driver
    -----------------

    Generally speaking drivers behave just like an instance of `file`, except
    they operate on data stored in a very specific way.  Drivers must handle file
    opening, closing, reading, and writing, and must pass an object that behaves
    like `file` to `BaseDriver.__init__()`.  The really critical methods are
    `__iter__()` and `write()`.  The former must yield one dictionary per
    iteration and the latter must accept a dictionary and write it to disk.

    See the `NewlineJSON()` driver for an example of a really simple driver
    and the `MsgPack()` driver for one that is more complex.
    """

    by_name = {}
    by_extension = {}
    register = False
    io_modes = ('r', 'w', 'a')

    def __init__(self, stream):

        """
        Creates an object that transparently interacts with all supported drivers
        by calling `stream`'s methods.

        Parameters
        ----------
        stream : <object>
            An object provided by a driver that behaves like `file`.
        """

        self._stream = stream

    def __repr__(self):
        return "<%s driver %s, mode '%s'>" % (
            'closed' if self.closed else 'open',
            self.driver_name,
            self.mode
        )

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()

    def __iter__(self):
        return iter(self.stream)

    @property
    def stream(self):

        """
        A handle to the underlying file-like object.  This can range from simply
        being an instance of `file()` to being an instance of a compression
        driver, which in turn has a `stream` property that is a driver class etc.
        """

        return self._stream

    def next(self):

        """
        This method must be explicitly defined, otherwise the `__getattr__()`
        method will return the `stream`'s `next()` method, which is not in the
        same namespace as `__iter__()` so this class will appear to have the
        necessary methods to be an iterator but will raise exceptions when
        iterated over.
        """

        return next(self.stream)

    __next__ = next

    def __getattr__(self, item):

        """
        For all other methods, just get it from the underlying `stream`.
        """

        return getattr(self.stream, item)


class NewlineJSON(BaseDriver):

    """
    Driver for accessing data stored as newline delimited JSON.
    """

    driver_name = 'NewlineJSON'
    extensions = ('json', 'nljson')

    def __init__(self, f, mode='r', **kwargs):

        """
        The object returned by `newlinejson.open()` has all of the methods
        required by `BaseDriver()` so this constructor is pretty sparse.  The
        `newlinejson.open()` function handles all of the file opening/closing
        and the returned object is passed directly to `BaseDriver()`.

        Parameters
        ----------
        f : str or `newlinejson.Stream()`
            Path to a file that can be opened by `newlinejson.open()` or a fully
            instantiated `newlinejson.Stream()` object.
        mode : str, optional
            I/O mode for `newlinejson.open()`.
        kwargs : **kwargs, optional
            Additional keyword arguments for `newlinejson.open()`.
        """

        BaseDriver.__init__(
            self,
            newlinejson.open(f, mode=mode, **kwargs)
        )


class MsgPack(BaseDriver):

    """
    Driver for accessing data stored as MsgPack.
    """

    driver_name = 'MsgPack'
    extensions = ('msg', 'msgpack')

    class _MsgPackWriter(object):

        """
        A helper class to give this driver a `write()` method.  MsgPack doesn't
        offer a file-like object for writing and expects the user to do this:

            >>> import msgpack
            >>> packer = msgpack.Packer()
            >>> with open('out.msg', 'w') as f:
            ...     f.write(packer.pack({'key': 'val'}))

        which doesn't fit into the `gpsdio` driver model.
        """

        def __init__(self, f, **kwargs):

            """
            Store the properties required to make this thing work.

            Parameters
            ----------
            f : file
                A file-like object open for writing.
            kwargs : **kwargs, optional
                Additional keyword arguments for `msgpack.Packer()`.
            """

            self._packer = msgpack.Packer(**kwargs)
            self._f = f

        def write(self, msg):

            """
            Pack and write data to the underlying file-like object.
            """

            self._f.write(self._packer.pack(msg))

        def __getattr__(self, item):

            """
            For all other methods default to the underlying file-like object.
            """

            return getattr(self._f, item)

    class _MsgPackReader(msgpack.Unpacker):

        """
        A helper class to make `msgpack.Unpacker()` behave similarly to `file`.
        """

        def __init__(self, f, **kwargs):

            """
            Instantiate `msgpack.Unpacker()` from an open file-like object.

            Parameters
            ----------
            f : file
                A file-like object open for reading.
            kwargs : **kwargs, optional
                Additional keyword arguments for `msgpack.Unpacker()`.
            """

            self._f = f
            msgpack.Unpacker.__init__(self, f, **kwargs)

        def __getattr__(self, item):

            """
            For all other methods default to the underlying file-like object.
            """

            return getattr(self._f, item)

    def __init__(self, f, mode='r', **kwargs):

        """
        Constructs the necessary objects for reading or writing and hands them
        off to `BaseDriver()`.  When reading `msgpack.Unpacker()` is used but
        when writing a special helper `_MsgPackWriter()` is used.

        Parameters
        ----------
        f : str or file
            Input file path or open file-like object.
        mode : str, optional
            Mode to open `f` with.
        kwargs : **kwargs, optional
            Additional keyword arguments for `msgpack.Unpacker()` or
            `msgpack.Unpacker()`.
        """

        if isinstance(f, six.string_types):
            self._f = codecs_open(f, mode=mode)
        else:
            self._f = f
        if mode == 'r':
            BaseDriver.__init__(
                self,
                self._MsgPackReader(self._f, **kwargs)
            )
        else:
            BaseDriver.__init__(
                self,
                self._MsgPackWriter(self._f, **kwargs)
            )


class BaseCompressionDriver(BaseDriver):

    """
    A slightly modified subclass of `BaseDriver()` to allow separation of normal
    drivers and compression drivers.
    """

    by_name = {}
    by_extension = {}
    register = False


class GZIP(BaseCompressionDriver):

    """
    Driver for accessing data stored as GZIP.
    """

    driver_name = 'GZIP'
    extensions = 'gz',

    def __init__(self, f, mode='r', **kwargs):

        """
        Creates an instance of `gzip.GzipFile()` and passes it to `BaseDriver()`.

        If `f` is a string `gzip.open()` is used, otherwise `gzip.GzipFile()` is
        instantiated directly.

        Parameters
        ----------
        f : str or file
            File path or open file-like object.
        mode : str, optional
            I/O mode for the `gzip` library.
        kwargs : **kwargs, optional
            Additional keyword arguments for `gzip.open()` or `gzip.GzipFile()`.
        """

        if isinstance(f, six.string_types):
            BaseDriver.__init__(
                self,
                gzip.open(f, mode=mode, **kwargs)
            )
        else:
            BaseDriver.__init__(
                self,
                gzip.GzipFile(fileobj=f, mode=mode, **kwargs)
            )


class BZ2(BaseCompressionDriver):

    """
    Driver for accessing data stored as BZIP2.
    """

    driver_name = 'BZ2'
    extensions = 'bz2',

    def __init__(self, f, mode='r', **kwargs):

        """
        All arguments are passed directly to `bz2.BZFile()`.

        Parameters
        ----------
        f : str or file
            File path or open file-like object.
        mode : str, optional
            I/O mode for the `bz2` library.
        kwargs : **kwargs, optional
            Additional keyword arguments for `bz2.BZFile()`.
        """

        BaseDriver.__init__(
            self,
            bz2.BZ2File(f, mode=mode, **kwargs)
        )



"""
Schema definitions and functions to transform rows
"""


import six

from ._schema_def import DATETIME_FORMAT
from ._schema_def import datetime2str
from ._schema_def import fields_by_msg_type
from ._schema_def import str2datetime
from ._schema_def import VERSIONS


CURRENT = VERSIONS[max(VERSIONS.keys())]
default = {
    field: CURRENT[field]['default'] for field in CURRENT.keys() if 'default' in CURRENT[field]
}

schema_import_functions = {
    field: CURRENT[field]['import'] for field in CURRENT if CURRENT[field].get('import', None) is not None
}
schema_export_functions = {
    field: CURRENT[field]['export'] for field in CURRENT if CURRENT[field].get('export', None) is not None
}
schema_types = {
    field: CURRENT[field]['type'] for field in CURRENT if CURRENT[field].get('type', None) is not None
}
schema_cast_functions = schema_types.copy()
schema_cast_functions.update(schema_import_functions)


def validate_msg(row, ignore_missing=False, skip_failures=False, schema=CURRENT):

    """
    Validate that a row contains all fields required by its type and that they
    are of the required types. Returns True if valid, False if fields are
    missing or of wrong type.
    """

    res = True

    def add_invalid(key, value):
        if skip_failures:
            if '__invalid__' not in row:
                row['__invalid__'] = {}
            row['__invalid__'][key] = value

    try:
        for key, value in row.items():
            if key == '__invalid__':
                continue
            try:
                if key not in schema:
                    continue
                fieldschema = schema[key]
                if value is None:
                    if not fieldschema.get('null', False):
                        add_invalid(key, ("null", row.pop(key)))
                        res = False
                else:
                    vt = type(value)
                    origt = t = fieldschema.get('type', str)

                    # Hack to allow both UTF-encoded str and unicode
                    # strings - this seems to be container dependent,
                    # and actually converting using import/export
                    # would be to expensive and this is generally not
                    # that usefull
                    if t is str or t is unicode:
                        t = six.string_types

                    # Hack to allow ints where floats should be used,
                    # as the container format might convert whole
                    # numbers into ints under our feet.
                    if t is float and vt in six.integer_types:
                        vt = float

                    if not issubclass(vt, t):
                        if isinstance(origt, (list, tuple)):
                            name = [item.__name__ for item in origt]
                        else:
                            name = origt.__name__
                        add_invalid(key, (name, row.pop(key)))
                        res = False
                    elif 'test' in fieldschema and not fieldschema['test'](value):
                        add_invalid(key, ('test failed', row.pop(key)))
                        res = False
            except Exception as e:
                add_invalid(key, (str(e), row.pop(key)))
                res = False

        if not ignore_missing:
            if 'type' not in row:
                add_invalid('__missing_keys__', ('type',))
                res = False
            else:
                default_keys = set(get_default_msg(int(row['type']), schema=schema, optional=False).keys())
                row_keys = set(row.keys())
                if len(default_keys - row_keys) != 0:
                    add_invalid('__missing_keys__', tuple(default_keys - row_keys))
                    res = False

    except Exception as e:
        add_invalid('__exception__', str(e))
        res = False

    return res

def complete_msg(msg, schema=CURRENT):
    """Add any fields that are not present with their default values"""

    res = get_default_msg(int(msg['type']), schema=schema)
    res.update(msg)

    return res


def strip_msg(msg, schema=CURRENT):
    """Remove any fields that do not belong"""

    res = {'type': msg['type']}
    msg_type = int(msg['type'])

    for key, value in msg.iteritems():
        if key in fields_by_msg_type[msg_type]:
            res[key] = value

    return res


def force_msg(msg, schema=CURRENT, keep_fields=False):

    """
    Make sure a msg has only valid fields by removing unrecognized fields and
    adding missing ones. Input msg must have a `type` field containing the
    message type.

    No field validation or type-casting is performed EXCEPT the `type`
    field is always forced to be an `int` in order to avoid unnecessarily
    raising an exception when a string value is encountered.  If the string
    cannot be cast to an `int` then an exception will be raised.

    Parameters
    ----------
    msg : dict
        Input msg - must contain a `type` key
    keep_fields : bool
        Keep extraneous fields
    schema : dict
        The schema definition to use

    Returns
    -------
    dict
        Input msg forced to a specific message type
    """

    if not keep_fields:
        msg = strip_msg(msg, schema=schema)
    return complete_msg(msg, schema=schema)

    message = get_default_msg(int(msg['type']), schema=schema)


def import_msg(row, skip_failures=False, cast_values=False):

    """
    Cast all values in a row from their import types as defined by the
    schema definition

    Parameters
    ----------
    row : dict
        Input row with normalized field names
    skip_failures: bool
        If False, reading a row with an attribute value that does not match
        the schema type for that attribute will cause an exception.
    cast_values: bool
        If true, an attempt will be made to cast values to the right types
        even for primitive types where the import type and the python type is the same.

    Returns
    -------
    dict
    """

    if cast_values:
        import_functions = schema_cast_functions
    else:
        import_functions = schema_import_functions

    # This function sometimes receives rows that are not exclusively AIS message types
    # The assumption is that the user wants them to be there since there are enough ways
    # to explicitly strip them off
    output = {}
    for field, val in six.iteritems(row):
        if field in import_functions:
            try:
                output[field] = import_functions[field](val)
            except Exception as e:
                if not skip_failures:
                    raise Exception("%s=%s: %s: %s" % (field, repr(val), type(e), e))
                if '__invalid__' not in output:
                    output['__invalid__'] = {}
                output['__invalid__'][field] = val
        else:
            output[field] = val

    return output


def export_msg(row, skip_failures=False):

    """
    Cast all values in a row to their export types as defined by the
    schema definition

    Parameters
    ----------
    row : dict
        Input row adhering to the GPSD schema
    skip_failures: bool
        If False, reading a row with an attribute value that does not match
        the schema type for that attribute will cause an exception.
    """

    # This function sometimes receives rows that are not exclusively AIS message types
    # The assumption is that the user wants them to be there since there are enough ways
    # to explicitly strip them off
    output = {}
    for field, val in row.items():
        if field in schema_export_functions:
            try:
                output[field] = schema_export_functions[field](val)
            except Exception as e:
                if not skip_failures:
                    raise Exception("%s: %s: %s" % (field, type(e), e))
                if '__invalid__' not in output:
                    output['__invalid__'] = {}
                output['__invalid__'][field] = str(val)
        else:
            output[field] = val

    return output


def get_default_msg(msg_type, schema=CURRENT, optional=True):

    """
    Get an AIS message containing nothing but default values

    Parameters
    ----------
    msg_type : int
        Message type ID
    schema : dict, optional
        Schema definition from which to extract default values
    """

    try:
        res = {field: schema[field]['default']
               for field in fields_by_msg_type[msg_type]
               if 'default' in schema[field] and (optional or schema[field].get('required', True))}
        res['type'] = msg_type
        return res
    except Exception as e:
        raise ValueError("Invalid AIS message type: %s: %s" % (msg_type, e))
